# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import os
import logging
import collections

# Avoid TensorFlow spam:
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
os.environ['CUDA_VISIBLE_DEVICES'] = '0'
logging.getLogger('tensorflow').addFilter(
    lambda record: 'Tracing is expensive and the excessive' not in record.msg
)

from . import pycompat
from . import utils
from . import constants
from . import sharknado # todo remove
# from . import gamey
# from . import commanding

__VersionInfo = collections.namedtuple('VersionInfo',
                                       ('major', 'minor', 'micro'))

__version__ = '0.5.1'
__version_info__ = __VersionInfo(*(map(int, __version__.split('.'))))

del os, logging, collections, __VersionInfo # Avoid polluting the namespace



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import click
import webbrowser
import pathlib
import itertools
import re
import time
from typing import (Optional, Tuple, Union, Container, Hashable, Iterator, Mapping,
                    Iterable, Any, Dict, FrozenSet, Callable, Type, Sequence, Set)
import logging
import platform
import sys
import contextlib

from marley.jamswank import jamming
from . import logging_setup
from . import gamey
from .constants import swank_database

logger = logging.getLogger(__name__)


@click.group()
@click.option('-v', '--verbose', is_flag=True)
@click.option('--log-to-file/--dont-log-to-file', is_flag=True, default=True)
def marley_command_group(*, verbose: bool = False, log_to_file: bool = True) -> None:
    from marley import __version__
    logging_setup.setup(verbose=verbose, log_to_file=log_to_file)
    logger.debug(f'Starting Marley {__version__}, Python version {platform.python_version()}')
    logger.debug(f'{sys.argv=}')
    gamey.global_gamey_swank_database = swank_database

@marley_command_group.resultcallback()
def marley_done(result: Any, *, verbose: bool = False, log_to_file: bool = True) -> None:
    logger.debug(f'Marley finished, exiting.')

@marley_command_group.command()
def initialize() -> None:
    pass

@click.option('--host', default=None)
@click.option('--port', default=None)
@marley_command_group.command()
def serve(*, host: Optional[str], port: Optional[str]) -> None:
    from . import arezzo
    with arezzo.ArezzoServerThread(host=host, port=port) as arezzo_server_thread:
        logger.info(f'Open {arezzo_server_thread.url} in your browser to access Marley.')
        while True:
            time.sleep(0.1)

####################################################################################################

def is_exit_0_exception(exception: Exception) -> bool:
    return (
        (isinstance(exception, SystemExit) and exception.code == 0) or
        (isinstance(exception, click.exceptions.Exit) and exception.exit_code == 0)
    )


@contextlib.contextmanager
def run_and_log_exception():
    try:
        yield
    except BaseException as base_exception:
        if not is_exit_0_exception(base_exception):
            logger.exception('Marley exited because of an exception.')
            raise SystemExit(1) from base_exception

def import_all_worlds():
    import marley.worlds.grid_royale
    import marley.worlds.blackjack


def marley(*args: Any, **kwargs: Any) -> None:
    # Import all the worlds to register their command groups:
    import_all_worlds()

    with run_and_log_exception():
        marley_command_group(*args, **kwargs)




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import pathlib
import os
import threading

import filelock

class InternalFileLock(filelock.FileLock):
    pass


class CachedFileLockType(type):
    def __new__(mcls, *args, **kwargs) -> CachedFileLockType:
        result = super().__new__(mcls, *args, **kwargs)
        result.__cache = {}
        return result

    def __call__(cls, path: os.PathLike) -> FileLock:
        path = pathlib.Path(path)
        try:
            return cls.__cache[path]
        except KeyError:
            cls.__cache[path] = file_lock = super().__call__(path)
            return file_lock




class FileLock(metaclass=CachedFileLockType):
    def __init__(self, path: os.PathLike) -> None:
        self.path = pathlib.Path(path)
        self.internal_file_lock = InternalFileLock(path)
        self.threading_lock = threading.RLock()
        self._ensured_folder_exists = False

    def acquire(self) -> None:
        self.threading_lock.acquire()
        if not self._ensured_folder_exists:
            self.path.parent.mkdir(parents=True, exist_ok=True)
        try:
            self.internal_file_lock.acquire()
        except:
            self.threading_lock.release()
            raise

    def release(self) -> None:
        self.internal_file_lock.release()
        self.threading_lock.release()

    def __enter__(self) -> FileLock:
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.release()

    def __repr__(self) -> str:
        return f'{type(self).__name__}({repr(self.path)})'





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict)
import collections


class CuteUserDict(collections.UserDict):

    def __repr__(self) -> str:
        return f'{type(self).__name__}({collections.UserDict.__repr__(self)})'






# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

from .base_field import BaseField
from .simple_field import SimpleField
from .swank_field import SwankField
from .parchment_field import ParchmentField
from .savvy_field import SavvyField


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import click
import webbrowser
import pathlib
import re
import time
from typing import (Optional, Tuple, Union, Container, Hashable, Iterator, Mapping,
                    Iterable, Any, Dict, FrozenSet, Callable, Type, Sequence, Set)
import logging
import platform
import sys

import marley
from marley.jamswank import jamming
from .constants import LETTERS, DEFAULT_BOARD_SIZE, DEFAULT_N_FOOD_TILES, DEFAULT_N_PLAYERS
from marley.constants import jam_file_database
from .core import State, Policy, Game
from . import utils
from . import server
from marley import gamey

logger = logging.getLogger(__name__)

@marley.commanding.marley_command_group.group(name='grid-royale')
def command_group() -> None:
    pass


@click.option('--board-size', type=int, default=DEFAULT_BOARD_SIZE)
@click.option('--n-players', type=int, default=DEFAULT_N_PLAYERS)
@click.option('--n-food-tiles', type=int, default=DEFAULT_N_FOOD_TILES)
@click.option('--allow-shooting/--no-shooting', default=True)
@click.option('--allow-walling/--no-walling', default=False)
@click.option('--pre-train/--dont-pre-train', default=True)
@click.option('--pre-train-n-games', type=int, default=10)
@click.option('--pre-train-n-phases', type=int, default=30)
@click.option('--browser/--no-browser', 'open_browser', default=True)
@click.option('--host', default=server.DEFAULT_HOST)
@click.option('--port', default=server.DEFAULT_PORT)
@click.option('--max-length', default=None, type=int)
@command_group.command()
def play(*, board_size: int, n_players: int, n_food_tiles: int, allow_shooting: bool,
         allow_walling: bool, pre_train: bool, pre_train_n_games: int, pre_train_n_phases: int,
         open_browser: bool, host: str, port: str, max_length: Optional[int] = None) -> None:
    with server.ServerThread(host=host, port=port, quiet=True) as server_thread:

        if open_browser:
            logger.info(f'Opening {server_thread.url} in your browser to view the game.')
            webbrowser.open_new(server_thread.url)
        else:
            logger.info(f'Open {server_thread.url} in your browser to view the game.')

        make_initial_state = lambda: (
            State.make_initial(
                n_players=n_players, board_size=board_size, allow_shooting=allow_shooting,
                allow_walling=allow_walling, n_food_tiles=n_food_tiles
            )
        )

        letters = LETTERS[:n_players]

        culture = gamey.Culture(
            player_id_to_policy={letter: Policy.create(board_size=board_size) for letter in
                                 letters[:n_players]}
        )

        if pre_train:
            culture = culture.train_progress_bar(
                make_initial_state, n_games=pre_train_n_games,
                max_game_length=30, n_phases=pre_train_n_phases
            )

        game: Game = Game.from_state_culture(make_initial_state(), culture)

        if max_length is None:
            logger.info(f'Calculating states in the simulation, press Ctrl-C to stop.')
        else:
            logger.info(f'Calculating {max_length} states, press Ctrl-C to stop.')

        jam_parchment = jam_file_database['games'][jamming.JamId.create(block_size=1_000)]
        for _state in game.write_to_jam_parchment(jam_parchment, max_length=max_length):
            pass
        logger.info(f'Finished calculating {max_length} states, still serving forever.')
        while True:
            time.sleep(0.1)


@click.option('--n-players', type=int, default=3)
@click.option('--n-food-tiles', type=int, default=5)
@click.option('--allow-shooting/--no-shooting', default=True)
@click.option('--allow-walling/--no-walling', default=False)
@click.option('--browser/--no-browser', 'open_browser', default=True)
@click.option('--host', default=server.DEFAULT_HOST)
@click.option('--port', default=server.DEFAULT_PORT)
@click.option('--max-length', default=300, type=int)
@command_group.command()
def demo(*, n_players: int, n_food_tiles: int, allow_shooting: bool, allow_walling: bool,
         open_browser: bool, host: str, port: str, max_length: Optional[int] = None) -> None:
    board_size = 6
    with server.ServerThread(host=host, port=port, quiet=True) as server_thread:

        if open_browser:
            logger.info(f'Opening {server_thread.url} in your browser to view the game.')
            webbrowser.open_new(server_thread.url)
        else:
            logger.info(f'Open {server_thread.url} in your browser to view the game.')

        make_initial_state = lambda: (
            State.make_initial(
                n_players=n_players, board_size=board_size, allow_shooting=allow_shooting,
                allow_walling=allow_walling, n_food_tiles=n_food_tiles
            )
        )

        letters = LETTERS[:n_players]

        serialized_model = (pathlib.Path(__file__).parent / 'demo.h5').read_bytes()
        culture = gamey.Culture(
            player_id_to_policy={
                letter: Policy.create(board_size=board_size, serialized_models=(serialized_model,))
                for letter in letters[:n_players]
            }
        )

        game: Game = Game.from_state_culture(make_initial_state(), culture)

        if max_length is None:
            logger.info(f'Calculating states in the simulation, press Ctrl-C to stop.')
        else:
            logger.info(f'Calculating {max_length} states, press Ctrl-C to stop.')

        jam_parchment = jam_file_database['games'][jamming.JamId.create(block_size=1_000)]
        for _state in game.write_to_jam_parchment(jam_parchment, max_length=max_length):
            pass
        logger.info(f'Finished calculating {max_length} states, still serving forever.')
        while True:
            time.sleep(0.1)



@click.option('--host', default=server.DEFAULT_HOST)
@click.option('--port', default=server.DEFAULT_PORT)
@command_group.command()
def serve(*, host: str, port: str) -> None:
    with server.ServerThread(host=host, port=port) as server_thread:
        logger.info(f'Open {server_thread.url} in your browser to view the game.')
        while True:
            time.sleep(0.1)



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import json
import math
import sys
import re
import time
import itertools
import numbers
import collections.abc
import urllib.parse
from typing import Union, Optional, Iterator, Iterable, TypeVar, Callable, Tuple, TypeVar, Sequence
import random

from browser import document, html, ajax, timer, window

def add_parameters_to_url(url: str, parameters: dict) -> str:
    # Todo: This is a caveman implementation, replace with urllib.parse
    if '?' in url:
        raise NotImplementedError
    return f'{url}?{"&".join(f"{key}={value}" for key, value in parameters.items())}'


def cool_ajax(url: str, callback: Callable, method: str = 'GET', *, mode: str = 'text',
              disable_cache: bool = True) -> None:
    assert mode in ('text', 'binary')
    request = ajax.ajax(mode=mode)
    if disable_cache:
        url = add_parameters_to_url(url, {'_': random.randint(0, 10**8)})
    request.open(method, url, True)
    if mode == 'binary':
        request.responseType = 'blob'
    request.bind('complete', callback)
    request.send()


def sliced(seq: Sequence, n: int) -> Iterable[Sequence]:
    """Yield slices of length *n* from the sequence *seq*.

        >>> list(sliced((1, 2, 3, 4, 5, 6), 3))
        [(1, 2, 3), (4, 5, 6)]

    If the length of the sequence is not divisible by the requested slice
    length, the last slice will be shorter.

        >>> list(sliced((1, 2, 3, 4, 5, 6, 7, 8), 3))
        [(1, 2, 3), (4, 5, 6), (7, 8)]

    This function will only work for iterables that support slicing.
    For non-sliceable iterables, see :func:`chunked`.

    """
    # Taken from more_itertools
    return itertools.takewhile(len, (seq[i : i + n] for i in itertools.count(0, n)))



# Copyright 2020 Ram Rachum and collaborators.,
# This program is distributed under the MIT license.

from __future__ import annotations
from typing import Iterable, Tuple, Any

import sys
import itertools
import pathlib
import random
import enum
import functools
import io
import numbers
import logging
from typing import Optional, Union
import contextlib
import shlex

import numpy as np
import click
import more_itertools
import numpy as np

from marley import gamey
from marley import jamswank
from marley import sharknado
from .core import *
from .sharknadoing import BlackjackProject

def get_chart_html(blackjack_project: BlackjackProject) -> str:
    import altair
    import pandas as pd

    with blackjack_project.parchment_lock:
        raw_data = tuple(
            tuple(game_sequence.mean_score for game_sequence in agent.game_sequences)
            for agent in blackjack_project.agents
        )
        baseline_data_frame = pd.DataFrame(
            (
                (baseline_policy_evaluation.title,
                 baseline_policy_evaluation.game_sequence.mean_score)
                for baseline_policy_evaluation in blackjack_project.baseline_policy_evaluations
            ),
            columns=('name', 'score')
        )

    data_frame = pd.DataFrame(
        (
            (i_policy, generation, score)
            for i_policy, scores in enumerate(raw_data)
            for generation, score in enumerate(scores)
        ),
        columns=('i_policy', 'generation', 'score')
    )


    base_chart = altair.Chart(data_frame)

    base_chart_with_generation = (
        base_chart.
        encode(
            altair.X(
                'generation', type='quantitative',
                axis=altair.Axis(title='generations', tickMinStep=1)
            ),
        )
    )


    score_mean_chart = (
        base_chart_with_generation.
        mark_line().
        encode(
            altair.Y('score', type='quantitative', aggregate='mean')
        )
    )

    score_ci_chart = (
        base_chart_with_generation.
        mark_errorband(extent='ci').
        encode(
            altair.Y('score', type='quantitative')
        )
    )

    baseline_rule_chart = (
        altair.Chart(baseline_data_frame).
        mark_rule().
        encode(
            altair.Y('score')
        )
    )

    baseline_text_chart = (
        altair.Chart(baseline_data_frame).
        mark_text(
            align='left',
            x='width',
            dx=5,
            fontSize=15,
        ).
        encode(
            altair.Y('score'),
            altair.Text('name')
        )
    )


    score_chart = altair.layer(score_mean_chart, score_ci_chart)
    baseline_chart = altair.layer(baseline_rule_chart, baseline_text_chart)

    chart_title = (f'The gradual rise of the score achieved by {len(blackjack_project.agents)} '
                   f'agents with learning policies, compared with a few baseline '
                   f'policies. The shaded area is a bootstrapped 95% confidence interval.')
    chart = (
        altair.layer(score_chart, baseline_chart, title=chart_title).
        properties(width=800, height=600)
    )


    string_io = io.StringIO()
    chart.save(string_io, format='html')
    return string_io.getvalue()




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import datetime as datetime_module
from typing import Iterable, Mapping, Union, Type, Optional, Any
import itertools
import collections.abc

from ...jamming import BaseJamDatabase, Jam, JamId, JamItem, JamKind, JamParchment
from ..core import Swank, SwankDatabase



class BaseField(abc.ABC):
    field_type_name: str

    @abc.abstractmethod
    def from_jam(self, jam: Jam, swank_database: SwankDatabase, *, swank_cache: dict) -> Swank:
        raise NotImplementedError

    @abc.abstractmethod
    def to_jam(self, value: Optional[Any], swank_database: SwankDatabase) -> Jam:
        raise NotImplementedError

    def get_default_value(self, swank_database: SwankDatabase) -> Any:
        return self.default_factory()

    default_factory = type(None)

    def __set_name__(self, swank_type: Type[Swank], name) -> None:
        self.name = name

    def __get__(self, swank: Optional[Swank], swank_type: Optional[Type] = None) -> Any:
        if swank is None:
            assert swank_type is not None
            return self
        return swank._Swank__field_values[self.name]

    def __set__(self, swank: Swank, value: Any) -> None:
        swank._Swank__field_values[self.name] = value

    def __delete__(self, swank: Swank) -> None:
        raise NotImplementedError

    def _reduce(self) -> tuple:
        return (type(self), self.name)

    def __eq__(self, other: Any) -> bool:
        return (type(other) == type(self)) and (self._reduce() == other._reduce())

    def __hash__(self) -> int:
        return hash(self._reduce())

    def __repr__(self) -> str:
        return f'{type(self).__name__}({repr(self.name)})'







# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import math
import inspect
import re
import abc
import random
import itertools
import collections.abc
import statistics
import concurrent.futures
import enum
import functools
import numbers
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar)
import dataclasses

import more_itertools
import numpy as np

from marley.utils import ImmutableDict
from marley.jamswank import JamId, SwankDatabase
from marley.jamswank.swanking import SimpleField, SavvyField, ParchmentField, SwankField
from . import utils
from . import exceptions
from . import aggregating
from .base import PlayerId, Story
from marley.gamey.gamey_swanking import GameySwank


class Narrative(collections.abc.Sequence):
    def __init__(self, game: Game, player_id: PlayerId) -> None:
        self.game = game
        self.player_id = player_id

    def __len__(self) -> int:
        return len(self.game) - 1

    def __getitem__(self, i: Union[int, slice]) -> int:
        if isinstance(i, slice):
            raise NotImplementedError
        assert isinstance(i, int)
        if - (len(self) + 1) < i < 0:
            i += len(self)
        if not 0 <= i <= len(self) - 1:
            raise IndexError
        return Story(
            old_observation=self.game.states[i][self.player_id],
            action=self.game.activities[i][self.player_id],
            reward=self.game.payoffs[i][self.player_id],
            new_observation=self.game.states[i + 1][self.player_id],
        )

    def __repr__(self):
        return (f'<{type(self).__name__} with {len(self)} stories>')


class NarrativeManager(collections.abc.Mapping):
    def __init__(self, game: Game) -> None:
        self.game = game

    def __len__(self) -> int:
        return len(self.game.culture)

    def __iter__(self) -> int:
        return iter(self.game.culture)

    def __getitem__(self, player_id: PlayerId) -> int:
        assert player_id in self.game.culture
        return Narrative(self.game, player_id)

    def __repr__(self):
        return (f'<{type(self).__name__} with {len(self)} narratives>')

class Game(collections.abc.Sequence, GameySwank):
    culture = SwankField()
    states = ParchmentField()
    activities = ParchmentField()
    payoffs = ParchmentField()
    def __init__(self, *, jam_id: Optional[Union[JamId, str]] = None,
                 jam_index: Optional[int] = None, swank_database: Optional[SwankDatabase] = None,
                 culture: aggregating.Culture, states: Iterable[aggregating.State],
                 activities: Iterable[aggregating.Activity],
                 payoffs: Iterable[aggregating.Payoff]) -> None:
        GameySwank.__init__(self, jam_id=jam_id, jam_index=jam_index, swank_database=swank_database,
                            culture=culture, states=states, activities=activities, payoffs=payoffs)
        self.narratives = NarrativeManager(self)
        self._assert_correct_lengths()

    def save(self) -> None:
        self.culture.save()
        GameySwank.save(self, all_parchment_fields=True)


    @classmethod
    def from_state_culture(cls, state: aggregating.State, culture: aggregating.Culture) -> Game:
        if not culture.has_jam_id_and_index:
            culture.save()
        return cls(culture=culture, states=(state,), activities=(), payoffs=())

    def _assert_correct_lengths(self) -> None:
        assert (len(self.states) == len(self.activities) + 1 == len(self.payoffs) + 1)


    def __iter__(self) -> Iterator[aggregating.State]:
        for states in Game.multi_iterate((self,)):
            (state,) = states
            yield state

    def __len__(self) -> int:
        return len(self.states)

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: len={len(self)}>'


    def __getitem__(self, i: Union[int, slice]) -> aggregating.State:
        if isinstance(i, slice):
            raise NotImplementedError
        else:
            assert isinstance(i, int)
            return self.states[i]



    def crunch(self, n: Optional[int] = None) -> None:
        for _ in more_itertools.islice_extended(self)[:n]:
            pass
        assert self.states[-1].is_end or len(self.states) == n

    @staticmethod
    def multi_crunch(games: Sequence[Game], n: Optional[int] = None) -> None:
        for _ in more_itertools.islice_extended(Game.multi_iterate(games))[:n]:
            pass


    @staticmethod
    def multi_iterate(games: Sequence[Game]) -> Iterator[tuple[Optional[aggregating.State], ...]]:
        from . import policing

        for game in games:
            game._assert_correct_lengths()

        finished_game_indices = set()

        def enumerate_unfinished_games():
            for i_game, game in enumerate(games):
                if i_game not in finished_game_indices:
                    yield (i_game, game)


        for i_state in itertools.count():
            state_deck = [None] * len(games)
            game_indices_to_play = []
            for i_game, game in enumerate_unfinished_games():
                try:
                    state_deck[i_game] = game.states[i_state]
                except IndexError:
                    game_indices_to_play.append(i_game)



            games_to_play = tuple(games[i_game] for i_game in game_indices_to_play)

            ### Making Q-policies way faster by batching matrix operations: ########################
            #                                                                                      #
            # This section is an optimization, it's a no-op when given policies that don't support
            # it.

            q_policy_to_observations = collections.defaultdict(list)
            iterator = itertools.chain.from_iterable(((game.states[-1] + game.culture).values()
                                                      for game in games_to_play))
            for observation, policy in iterator:
                if isinstance(policy, policing.QPolicy):
                    q_policy_to_observations[policy].append(observation)

            for q_policy, observations in q_policy_to_observations.items():
                q_policy.get_qs_for_observations(observations) # Automatically caches the result.

            #                                                                                      #
            ### Finished making Q-policies way faster by batching matrix operations. ###############

            for i_game, game in zip(game_indices_to_play, games_to_play):
                assert state_deck[i_game] is None
                state: aggregating.State = game.states[-1]

                if state.is_end:
                    finished_game_indices.add(i_game)
                    continue

                activity = game.culture.get_next_activity(state)
                game.activities.append(activity)

                next_payoff, next_state = state.get_next_payoff_and_state(activity)
                game.payoffs.append(next_payoff)
                game.states.append(next_state)

                state_deck[i_game] = next_state

                game._assert_correct_lengths()


            if all((state is None) for state in state_deck):
                assert finished_game_indices == set(range(len(games)))
                return
            yield tuple(state_deck)


    @property
    def total_payoff(self, *, start: int = 0, discount: numbers.Real = 1):
        assert start >= 0
        assert 0 <= discount <= 1
        player_id_to_reward = {}
        payoffs = more_itertools.islice_extended(self.payoffs)[start:]
        for i, payoff in enumerate(payoffs):
            for player_id, reward in payoff.items():
                player_id_to_reward[player_id] = (player_id_to_reward.get(player_id, 0) +
                                                   (discount ** i) * reward)

        return aggregating.Payoff(player_id_to_reward=player_id_to_reward)






# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.


from .pointing import Point
from .point_containing import PointContainer
from .intervaling import Interval
from .int_crowding import IntCrowd


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import os
import collections
import logging

from . import utils
from .core import *
from .commanding import command_group




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

'''Entry point for starting GridRoyale.'''

import logging
import contextlib

import click

logger = logging.getLogger(__name__)


if __name__ == '__main__':
    import marley.commanding
    marley.commanding.marley()


# Copyright 2020 Ram Rachum and collaborators.,
# This program is distributed under the MIT license.

from __future__ import annotations
from typing import Iterable, Tuple, Any

import sys
import itertools
import pathlib
import random
import enum
import functools
import io
import numbers
import logging
from typing import Optional, Union
import contextlib
import shlex

import numpy as np
import click
import more_itertools
import numpy as np

from marley import gamey
from marley import jamswank
from marley import sharknado
from .core import *

N_GAMES_PER_SEQUENCE = 300
N_GAMES_PER_BASELINE_POLICY_EVALUATION = 3_000

BASELINE_POLICY_TYPES_AND_ARGS = (
    (AlwaysStickPolicy, ()),
    (ThresholdPolicy, (15,)),
    (ThresholdPolicy, (16,)),
)


class BlackjackProject(gamey.GameySwank):
    def get_job(self, n_generations: int = 3, n_agents: int = 2) -> BlackjackJob:
        return BlackjackJob(self, n_generations=n_generations, n_agents=n_agents)

    agents = jamswank.ParchmentField()
    baseline_policy_evaluations = jamswank.ParchmentField()


    def get_chart_html(self) -> str:
        from .charting import get_chart_html
        return get_chart_html(self)

    def write_chart_to_desktop(self) -> None:
        chart_path = pathlib.Path.home() / 'Desktop' / 'chart.html'
        chart_path.write_text(self.get_chart_html())
        logger.info(f'Wrote chart to {shlex.quote(str(chart_path))}')



class Agent(gamey.GameySwank):
    policies = jamswank.ParchmentField()
    game_sequences = jamswank.ParchmentField()


class GameSequence(gamey.GameySwank):
    games = jamswank.ParchmentField()
    mean_score = jamswank.SimpleField()

    def defrag(self):
        first_game = self.games[0]
        # swank_database = first_game.swank_database
        parchment_field_names = ('states', 'activities', 'payoffs')
        parchments = {
            parchment_field_name: getattr(first_game, parchment_field_name).jam_parchment
            for parchment_field_name in parchment_field_names
        }
        for parchment_field_name, parchment in parchments.items():
            getattr(first_game, parchment_field_name).end_index = len(parchment)
        first_game.save()
        # culture_jam_item = swank_database.get_jam_item(first_game.culture)
        # assert culture_jam_item.jam_index == 0
        # cultures_parchment: jamswank.JamParchment = culture_jam_item.jam_parchment
        # assert len(cultures_parchment) == 1

        for i, (left_game, right_game) in enumerate(gamey.utils.iterate_windowed_pairs(self.games)):
            for parchment_field_name in parchment_field_names:
                left_game_parchment_ref = getattr(left_game, parchment_field_name)
                right_game_parchment_ref = getattr(right_game, parchment_field_name)
                right_game_parchment_ref.migrate(
                    left_game_parchment_ref.jam_id,
                    left_game_parchment_ref.end_index,
                    left_game_parchment_ref.end_index + len(right_game_parchment_ref),
                )
            # old_culture_parchment = swank_database.get_jam_item(right_game.culture).jam_parchment
            # right_game.culture.jam_id = cultures_parchment.jam_id
            # right_game.culture.jam_index = i + 1
            right_game.save()
            # old_culture_parchment.delete()



class BaselinePolicyEvaluation(gamey.GameySwank):
    game_sequence = jamswank.SwankField()
    title = jamswank.SimpleField()



class JobMixin:

    def __init__(self,
                 blackjack_project_or_ref: Union[BlackjackProject, jamswank.SwankRef]) -> None:
        self.blackjack_project_ref = jamswank.SwankRef.from_swank_or_ref(blackjack_project_or_ref)

    def _reduce(self) -> tuple:
        return (type(self), self.blackjack_project_ref)

    def get_agent_ref(self, i_agent: int) -> jamswank.SwankRef:
        with self.blackjack_project_ref.lock_and_load() as blackjack_project:
            try:
                agent = blackjack_project.agents[i_agent]
            except IndexError:
                blackjack_project.agents[i_agent] = agent = Agent()
                agent.save()
                blackjack_project.save()
            return jamswank.SwankRef.from_swank_or_ref(agent)




class BlackjackJob(JobMixin, sharknado.ThinJob):
    def __init__(self, blackjack_project: BlackjackProject, n_generations: int = 5,
                 n_agents: int = 4) -> None:
        JobMixin.__init__(self, blackjack_project)
        self.n_generations = n_generations
        self.n_agents = n_agents

    def _reduce(self) -> tuple:
        return (type(self), self.blackjack_project_ref, self.n_generations, self.n_agents)

    def get_parent_job_to_weight(self):
        return {
            AgentJob(self.blackjack_project_ref, n_generations=self.n_generations):
                                                             sharknado.TadpoleWeight(self.n_agents),
            EvaluateBaselinePolicyJob(self.blackjack_project_ref):
                                       sharknado.TadpoleWeight(len(BASELINE_POLICY_TYPES_AND_ARGS)),
        }

    def thin_run(self):
        with self.blackjack_project_ref.lock_and_load() as blackjack_project:
            blackjack_project.write_chart_to_desktop()



class AgentJob(JobMixin, sharknado.ParallelJob):
    dimensions = 1
    def __init__(self,
                 blackjack_project_or_ref: Union[BlackjackProject, jamswank.SwankRef], *,
                 n_generations: int) -> None:
        JobMixin.__init__(self, blackjack_project_or_ref)
        self.n_generations = n_generations


    def get_parent_job_to_weight(self):
        return {
            PolicyJob(self.blackjack_project_ref):
                sharknado.TadpoleWeight(self.n_generations),
            AgentGameSequenceJob(self.blackjack_project_ref):
                sharknado.TadpoleWeight(self.n_generations),
        }


class AgentGameSequenceJob(JobMixin, sharknado.ParallelJob):
    dimensions = 2

    def fat_sniff(self, fat_gain: sharknado.FatGain) -> sharknado.IntCrowd:
        agent_int_crowds = fat_gain.int_crowd.separate((0,))

        i_agents = tuple(more_itertools.first(agent_int_crowd)[0]
                         for agent_int_crowd in agent_int_crowds)
        if not i_agents:
            return
        i_to_agent = {}
        result_int_crowds = []

        with self.blackjack_project_ref.lock_and_load() as blackjack_project:
            for i_agent in i_agents:
                try:
                    agent = blackjack_project.agents[i_agent]
                except IndexError:
                    continue
                else:
                    i_to_agent[i_agent] = agent

        if i_to_agent:
            arbitrary_agent = more_itertools.first(i_to_agent.values())
            with arbitrary_agent.parchment_lock:
                for i_agent, agent_int_crowd in zip(i_agents, agent_int_crowds):
                    try:
                        agent = i_to_agent[i_agent]
                    except KeyError:
                        continue
                    result_int_crowds.append(
                        sharknado.IntCrowd(
                            sharknado.Point((i_agent, i_game_sequence))
                            for (i_agent, i_game_sequence) in agent_int_crowd
                            if agent.game_sequences.has_index(i_game_sequence)
                        )
                    )

        return sharknado.utils.union(result_int_crowds) if result_int_crowds else None

    def fat_run(self, fat_gain: sharknado.FatGain) -> None:
        (i_agent, i_policy) = fat_gain.int_crowd.get_single_point()
        agent_ref = self.get_agent_ref(i_agent)
        with agent_ref.lock_and_load() as agent:
            policy: ModelFreeLearningPolicy = agent.policies[i_policy]
        game_sequence = GameSequence()
        game_sequence.save() # Saving it separately, so we add it to the project only after it's
                             # done, and never have to think about partially-written game
                             # sequencees.
        culture = policy.make_culture()
        culture.save()
        for _ in range(N_GAMES_PER_SEQUENCE):
            game = gamey.Game.from_state_culture(BlackjackState.make_initial(), culture)
            game_sequence.games.append(game)

        gamey.Game.multi_crunch(game_sequence.games)

        game_sequence.mean_score = np.mean(
            tuple(
                sum(payoff.get_single() for payoff in game.payoffs) for game in game_sequence.games
            )
        )
        game_sequence.save()
        game_sequence.defrag()
        with agent_ref.lock_and_load(save=True) as agent:
            agent.game_sequences[i_policy] = game_sequence

    def get_parent_job_to_weight(self):
        return {PolicyJob(self.blackjack_project_ref): sharknado.CalfWeight()}


class PolicyJob(JobMixin, sharknado.SerialJob):
    dimensions = 2

    def fat_sniff(self, fat_gain: sharknado.FatGain) -> sharknado.IntCrowd:
        agent_int_crowds = fat_gain.int_crowd.separate((0,))

        i_agents = tuple(more_itertools.first(agent_int_crowd)[0]
                         for agent_int_crowd in agent_int_crowds)
        if not i_agents:
            return
        i_to_agent = {}
        result_int_crowds = []

        with self.blackjack_project_ref.lock_and_load() as blackjack_project:
            for i_agent in i_agents:
                try:
                    agent = blackjack_project.agents[i_agent]
                except IndexError:
                    continue
                else:
                    i_to_agent[i_agent] = agent

        if i_to_agent:
            arbitrary_agent = more_itertools.first(i_to_agent.values())
            with arbitrary_agent.parchment_lock:
                for i_agent, agent_int_crowd in zip(i_agents, agent_int_crowds):
                    try:
                        agent = i_to_agent[i_agent]
                    except KeyError:
                        continue
                    result_int_crowds.append(
                        sharknado.IntCrowd(
                            sharknado.Point((i_agent, i_policy)) for (i_agent, i_policy)
                            in agent_int_crowd if agent.policies.has_index(i_policy)
                        )
                    )

        return sharknado.utils.union(result_int_crowds) if result_int_crowds else None

    def fat_run(self, fat_gain: sharknado.FatGain) -> None:
        (i_agent, i_policy) = fat_gain.int_crowd.get_single_point()
        agent_ref = self.get_agent_ref(i_agent)

        if i_policy == 0:
            policy = ModelFreeLearningPolicy.create(discount=1)
        else:
            with agent_ref.lock_and_load() as agent:
                old_policy: ModelFreeLearningPolicy = agent.policies[i_policy - 1]
                game_sequence = agent.game_sequences[i_policy - 1]
            policy = old_policy.train(tuple(game.narratives[None] for game in game_sequence.games),
                                      n_epochs=5)

        with agent_ref.lock_and_load(save=True) as agent:
            agent.policies[i_policy] = policy

    def get_parent_job_to_weight(self):
        return {AgentGameSequenceJob(self.blackjack_project_ref): sharknado.CalfWeight(-1)}


class EvaluateBaselinePolicyJob(JobMixin, sharknado.ParallelJob):
    dimensions = 1

    def fat_sniff(self, fat_gain: sharknado.FatGain) -> sharknado.IntCrowd:
        policy_indicies = tuple(itertools.chain.from_iterable(fat_gain.int_crowd))
        assert all((0 <= policy_index < len(BASELINE_POLICY_TYPES_AND_ARGS))
                   for policy_index in policy_indicies)
        with self.blackjack_project_ref.lock_and_load() as blackjack_project:
            finished_policy_indices = tuple(
                filter(blackjack_project.baseline_policy_evaluations.has_index, policy_indicies)
            )
        return sharknado.IntCrowd(sharknado.Point((finished_policy_index,)) for
                                  finished_policy_index in finished_policy_indices)


    def fat_run(self, fat_gain: sharknado.FatGain) -> None:
        (policy_index,) = fat_gain.int_crowd.get_single_point()
        (baseline_policy_type, baseline_policy_args) = BASELINE_POLICY_TYPES_AND_ARGS[policy_index]
        baseline_policy: BlackjackPolicy = baseline_policy_type(*baseline_policy_args)

        game_sequence = GameSequence()
        game_sequence.save()

        baseline_policy_evaluation = BaselinePolicyEvaluation(
            title=baseline_policy.title,
            game_sequence=game_sequence,
        )
        baseline_policy_evaluation.save()

        culture = baseline_policy.make_culture()
        culture.save()
        for _ in range(N_GAMES_PER_BASELINE_POLICY_EVALUATION):
            game = gamey.Game.from_state_culture(BlackjackState.make_initial(), culture)
            game_sequence.games.append(game)

        gamey.Game.multi_crunch(game_sequence.games)

        game_sequence.mean_score = np.mean(
            tuple(
                sum(payoff.get_single() for payoff in game.payoffs)
                for game in game_sequence.games
            )
        )
        game_sequence.save()
        game_sequence.defrag()
        with self.blackjack_project_ref.lock_and_load(save=True) as blackjack_project:
            blackjack_project: BlackjackProject
            blackjack_project.baseline_policy_evaluations[policy_index] = baseline_policy_evaluation



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

'''
This module defines the GridRoyale game and all its rules.

# GridRoyale game rules #

Here are the rules of the game in simple form. For all the gritty details, check out the
`get_next_state_from_actions` method.

The world is a 2-dimensional grid, typically 20x20 in size. Each player is on a cell in the grid.
There can't be more than one player in a cell.

Each player basically wants to eat as much food as possible, avoid getting shot, and avoid colliding
with other players, walls and the edge of the board.

On each turn, each player does an action. It can be either moving, shooting or building a wall
(a.k.a "walling").

**Moving**: Moving is essential because it lets you walk towards food and away from bullets.

When a player is moving, it sometimes collides with a player, a wall or the edge of the
board. It collides with another player if that other player is trying to move into the same cell, or
is in the same cell and not moving. If there's a collision, the player's move is unsuccessful. It
remains in its old position, and it takes a punishment, currently -5 points.

If the move is successful, *and* the new cell has food, the player eats the food and gains 10
points. (Assuming it didn't get shot, more about that later.) The food will be removed from the
game, and a new piece of food will be spawned in a random free space on the board.

**Shooting**: A player may shoot a bullet in any of the four directions. The bullet will continue
travelling until it either hits a player, giving it a punishment of -10, or disappears off the edge
of the board, or hits a wall and destroys it. Shooting is useful to scare off other creatures who
might otherwise eat your food-- Assuming they're intelligent enough to avoid bullets.

Multiple bullets can freely pass through each other on their trajectory. They don't hit each other.

**Walling**: A player may build a wall in any of the four directions. A wall is a barrier that
blocks any player from entering that cell. The only way to remove a wall is to shoot it. Walls are
useful to prevent another player from moving into a certain area, especially if multiple walls are
built next to each other.
'''


from __future__ import annotations

import functools
import json
import pathlib
import string as string_module
import statistics
import webbrowser
import time
import itertools
import random
import time
import logging
import numbers
import io
import collections
from typing import (Optional, Tuple, Union, Container, Hashable, Iterator, Mapping,
                    Iterable, Any, Dict, FrozenSet, Callable, Type, Sequence, Set)
import contextlib
import re
import dataclasses
import datetime as datetime_module
import click
import abc

import numpy as np
import scipy.special
import more_itertools

from marley.jamswank import jamming, JamId
from marley.jamswank.swanking import SwankDatabase, SimpleField, SavvyField
from . import utils
from marley import gamey
from .utils import zip
from marley.utils import ImmutableDict
from .vectoring import Vector, Step, Position, Translation, Vicinity

from .constants import (SHOT_REWARD, COLLISION_REWARD, STARTING_REWARD, NOTHING_REWARD,
                        FOOD_REWARD, VISION_RANGE, VISION_SIZE, LETTERS, DEFAULT_BOARD_SIZE,
                        DEFAULT_N_FOOD_TILES, DEFAULT_N_PLAYERS)
from marley.constants import config_path, marley_folder

_action_neural_cache = {}


@functools.cache
def get_games_folder() -> pathlib.Path:
    config = read_config()
    try:
        games_folder_string = config['games_folder']
    except KeyError:
        return marley_folder / 'games'
    games_folder = pathlib.Path(games_folder_string)
    assert games_folder.is_dir()
    return games_folder


@dataclasses.dataclass(order=True, frozen=True)
class Action(gamey.Action):
    move: Optional[Step]
    shoot: Optional[Step]
    wall: Optional[Step]

    def __post_init__(self):
        if tuple(self).count(None) != 2:
            raise ValueError

    __iter__ = lambda self: iter((self.move, self.shoot, self.wall))


    def _to_neural(self) -> np.ndarray:
        result = np.zeros(12, dtype=bool)
        result[Action.all_actions.index(self)] = True
        return result

    def to_neural(self) -> np.ndarray:
        return _action_neural_cache[self]

    @property
    def name(self) -> str:
        if self.move is not None:
            return self.move.name
        elif self.shoot is not None:
            return f'shoot_{self.shoot.name}'
        else:
            return f'wall_{self.wall.name}'



(Action.up, Action.right, Action.down, Action.left) = \
    Action.all_move_actions = (Action(Step.up, None, None), Action(Step.right, None, None),
                                   Action(Step.down, None, None), Action(Step.left, None, None))

(Action.shoot_up, Action.shoot_right, Action.shoot_down, Action.shoot_left) = \
    Action.all_shoot_actions = (Action(None, Step.up, None), Action(None, Step.right, None),
                                   Action(None, Step.down, None), Action(None, Step.left, None))

(Action.wall_up, Action.wall_right, Action.wall_down, Action.wall_left) = \
    Action.all_wall_actions = (Action(None, None, Step.up), Action(None, None, Step.right),
                                   Action(None, None, Step.down), Action(None, None, Step.left))

Action.all_actions = Action.all_move_actions + Action.all_shoot_actions + Action.all_wall_actions

for action in Action:
    _action_neural_cache[action] = action._to_neural()


@dataclasses.dataclass(order=True, frozen=True)
class Bullet:
    '''A bullet that's shot by one player and might hit another player.'''
    position: Position
    direction: Step

    def get_next_bullet(self):
        return Bullet(position=(self.position + self.direction), direction=self.direction)

    def get_previous_bullet(self):
        return Bullet(position=(self.position - self.direction), direction=self.direction)




class Grid:
    '''Base class that represents a 2-dimensional square grid.'''
    board_size: int
    def __contains__(self, position: Position) -> bool:
        return 0 <= min(position) <= max(position) <= self.board_size - 1

    @staticmethod
    def iterate_random_positions(board_size: int) -> Iterator[Position]:
        while True:
            yield Position(
                random.randint(0, board_size - 1),
                random.randint(0, board_size - 1),
            )






class Observation(Grid, gamey.Observation):

    legal_move_actions = Action.all_move_actions

    def __init__(self, state: Optional[State], position: Position,
                 letter: str, score: int, last_action: Optional[Action],
                 reward: int = STARTING_REWARD) -> None:
        assert len(letter) == 1
        self.state = state
        self.position = position
        self.letter = letter.upper()
        self.score = score
        self.reward = reward
        self.last_action = last_action

    def to_savvy_content(self) -> tuple:
        return (self.state, self.position, self.letter, self.score, self.last_action, self.reward)

    @property
    def board_size(self) -> int:
        return self.state.board_size

    @property
    def legal_actions(self):
        return (
            *Action.all_move_actions,
            *(Action.all_shoot_actions if self.state.allow_shooting else ()),
            *(Action.all_wall_actions if self.state.allow_walling else ()),
        )

    @staticmethod
    def get_neural_dtype_for_board_size(board_size: int) -> np.dtype:
        return np.dtype(
            [('grid_input', bool, (board_size, board_size, 10)),
             ('sequential_input', bool, (len(Action),))]
        )

    @property
    def neural_dtype(self):
        return self.get_neural_dtype_for_board_size(self.board_size)

    def _to_neural(self) -> np.ndarray:

        array = np.zeros((1,), dtype=self.neural_dtype)
        grid_array = array[0]['grid_input']
        sequential_input_array = array[0]['sequential_input']

        ### Calculating grid subarray: #############################################################
        #                                                                                          #
        relative_player_position = Position(self.board_size // 2, self.board_size // 2)
        translation = relative_player_position - self.position

        for relative_position in Position.iterate_all(self.board_size):
            absolute_position: Position = relative_position - translation
            if absolute_position in self.state.food_positions:
                grid_array[tuple(relative_position) + (1,)] = True
            if absolute_position not in self:
                grid_array[tuple(relative_position) + (0,)] = True
            elif (bullets := self.state.bullets.get(absolute_position, None)):
                for bullet in bullets:
                    grid_array[tuple(relative_position) + (2 + bullet.direction.index,)] = True
            elif absolute_position in self.state.living_wall_positions:
                grid_array[tuple(relative_position) + (6,)] = True
            elif absolute_position in self.state.destroyed_wall_positions:
                grid_array[tuple(relative_position) + (7,)] = True
            elif (letter := self.state.position_to_letter.get(absolute_position, None)):
                if letter == self.letter:
                    grid_array[tuple(relative_position) + (8,)] = True
                else:
                    grid_array[tuple(relative_position) + (9,)] = True
        #                                                                                          #
        ### Finished calculating grid subarray. ####################################################

        ### Calculating sequential input subarray: #################################################
        #                                                                                          #
        last_action_neurals = (self.last_action.to_neural() if self.last_action is not None
                               else np.zeros(len(Action)))
        sequential_input_array[:] = np.concatenate((
            last_action_neurals,
        ))
        #                                                                                          #
        ### Finished calculating sequential input subarray. ########################################

        return array



    def p(self) -> None:
        print(self.state.ascii)


class State(Grid, gamey.State):

    default_block_size = 4_000

    Observation = Observation
    Action = Action
    is_end = False

    board_size = SimpleField()
    food_positions = SavvyField()
    allow_shooting = SimpleField()
    allow_walling = SimpleField()
    bullets = SavvyField()
    living_wall_positions = SavvyField()
    destroyed_wall_positions = SavvyField()

    def __init__(self, *, board_size: int, player_id_to_observation: Mapping[str, Observation],
                 food_positions: FrozenSet[Position], allow_shooting: bool = True,
                 allow_walling: bool = True,
                 bullets: ImmutableDict[Position, FrozenSet[Bullet]] = ImmutableDict(),
                 living_wall_positions: FrozenSet[Position],
                 destroyed_wall_positions: FrozenSet[Position],
                 jam_id: Optional[JamId] = None, jam_index: Optional[int] = None,
                 swank_database: Optional[SwankDatabase] = None) -> None:
        gamey.State.__init__(
            self, board_size=board_size, player_id_to_observation=player_id_to_observation,
            food_positions=food_positions, allow_shooting=allow_shooting,
            allow_walling=allow_walling, bullets=bullets,
            living_wall_positions=living_wall_positions,
            destroyed_wall_positions=destroyed_wall_positions, jam_id=jam_id, jam_index=jam_index,
            swank_database=swank_database
        )
        assert all(self.bullets.values()) # No empty sets in this bad boy.
        self.all_bullets = frozenset(itertools.chain.from_iterable(bullets.values()))
        self.position_to_letter = ImmutableDict(
            {observation.position: letter for letter, observation in self.items()}
        )
        self.wall_positions = living_wall_positions | destroyed_wall_positions

    def _reduce(self) -> tuple:
        return (
            type(self),
            frozenset(
                (letter, observation.position, observation.score, observation.reward,
                 observation.last_action) for letter, observation in self.items()
            ),
            self.bullets, self.food_positions, self.board_size, self.living_wall_positions,
            self.destroyed_wall_positions
        )

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, State) and self._reduce() == other._reduce()

    def __hash__(self) -> int:
        return hash(self._reduce())


    @staticmethod
    def make_initial(*, n_players: int = DEFAULT_N_PLAYERS, board_size: int = DEFAULT_BOARD_SIZE,
                     starting_score: int = 0, allow_shooting: bool = True,
                     allow_walling: bool = True, n_food_tiles: int = DEFAULT_N_FOOD_TILES) -> State:
        assert 1 <= n_players <= len(LETTERS)

        random_positions_firehose = utils.iterate_deduplicated(
                                     State.iterate_random_positions(board_size=board_size))
        random_positions = tuple(
            more_itertools.islice_extended(
                random_positions_firehose)[:(n_players + n_food_tiles)]
        )

        player_positions = random_positions[:n_players]
        food_positions = frozenset(random_positions[n_players:])
        assert len(food_positions) == n_food_tiles

        player_id_to_observation = {}
        for letter, player_position in zip(LETTERS, player_positions):
            player_id_to_observation[letter] = Observation(state=None, position=player_position,
                                                        score=starting_score, letter=letter,
                                                        last_action=None)

        state = State(
            board_size=board_size,
            player_id_to_observation=player_id_to_observation,
            food_positions=food_positions,
            allow_shooting=allow_shooting,
            allow_walling=allow_walling,
            living_wall_positions=frozenset(),
            destroyed_wall_positions=frozenset(),
        )

        for observation in player_id_to_observation.values():
            observation.state = state

        return state


    def get_next_payoff_and_state(self, activity: gamey.Activity) -> Tuple[gamey.Payoff, State]:
        new_player_position_to_olds = collections.defaultdict(set)
        wip_living_wall_positions = set(self.living_wall_positions)
        wip_destroyed_wall_positions = set() # Ignoring old destroyed walls, they're gone.

        for letter, action in activity.items():
            action: Action
            old_observation = self[letter]
            assert action in old_observation.legal_actions
            old_player_position = old_observation.position
            if action.move:
                new_player_position_to_olds[old_player_position +
                                                               action.move].add(old_player_position)
            else:
                new_player_position_to_olds[old_player_position].add(old_player_position)
                if action.wall and (wall_position := old_player_position + action.wall) in self:
                    wip_living_wall_positions.add(wall_position)


        ############################################################################################
        ### Figuring out which players collided: ###################################################
        #                                                                                          #
        # There are four types of collisions:
        # 1. A player trying to move out of the board.
        # 2. A player trying to move to a wall, living or destroyed.
        # 3. Two or more players that try to move into the same position.
        # 4. Two players that are trying to move into each other's positions.
        # 5. Any players that are trying to move into the old position of a player that had one
        #    of the two collisions above, and is therefore still occupying that position.

        collided_player_positions = set()


        while True:
            for new_player_position, old_player_positions in new_player_position_to_olds.items():
                if (new_player_position not in self) or \
                                                 (new_player_position in wip_living_wall_positions):
                    # This is either a type 1 or a type 2 collision.
                    collided_player_positions |= old_player_positions
                    del new_player_position_to_olds[new_player_position]
                    for old_player_position in old_player_positions:
                        new_player_position_to_olds[old_player_position].add(old_player_position)
                        if old_player_position in wip_living_wall_positions:
                            wip_living_wall_positions.remove(old_player_position)

                    # We modified the dict while iterating, let's restart the loop:
                    break

                if len(old_player_positions) >= 2:
                    # This is either a type 3 or a type 5 collision.
                    collided_player_positions |= old_player_positions
                    del new_player_position_to_olds[new_player_position]
                    for old_player_position in old_player_positions:
                        new_player_position_to_olds[old_player_position].add(old_player_position)

                    # We modified the dict while iterating, let's restart the loop:
                    break

                if (len(old_player_positions) == 1 and
                    ((old_player_position := more_itertools.one(old_player_positions)) !=
                      new_player_position) and new_player_position_to_olds.get(
                                               old_player_position, None) == {new_player_position}):
                    # Type 4 collision.
                    collided_player_positions |= {old_player_position, new_player_position}
                    new_player_position_to_olds[new_player_position] = {new_player_position}
                    new_player_position_to_olds[old_player_position] = {old_player_position}

                    # We modified the dict while iterating, let's restart the loop:
                    break

            else:
                # We already found all collisions, if any.
                break
        #                                                                                          #
        ### Finished figuring out which players collided. ##########################################
        ############################################################################################

        new_player_position_to_old = {
            new_player_position: more_itertools.one(old_player_positions) for
            new_player_position, old_player_positions in new_player_position_to_olds.items()
        }
        del new_player_position_to_olds # Prevent confusion

        # Disallowing building walls where there is a player that didn't move:
        for living_wall_position in tuple(wip_living_wall_positions):
            try:
                old_player_position = new_player_position_to_old[living_wall_position]
            except KeyError:
                pass
            else:
                assert old_player_position == living_wall_position
                wip_living_wall_positions.remove(living_wall_position)

        ############################################################################################
        ### Figuring out bullets: ##################################################################
        #                                                                                          #

        # Todo: This section needs a lot of tests!

        wip_bullets: Mapping[Position, Set[Bullet]] = collections.defaultdict(set)

        # Continuing trajectory for existing bullets:
        for bullet in self.all_bullets:
            new_bullet = bullet.get_next_bullet()
            wip_bullets[new_bullet.position].add(new_bullet)

        # Processing new bullets that were just fired:
        for letter, action in activity.items():
            if action.shoot is not None:
                player_position = self[letter].position
                new_bullet = Bullet(player_position + action.shoot, action.shoot)
                wip_bullets[new_bullet.position].add(new_bullet)

        # Clearing bullets out of board:
        for position in [position for position, bullets in wip_bullets.items()
                         if (position not in self)]:
            del wip_bullets[position]

        # Figuring out which walls were shot, removing these bullets:
        for living_wall_position in tuple(wip_living_wall_positions):
            if wip_bullets.pop(living_wall_position, None):
                wip_living_wall_positions.remove(living_wall_position)
                wip_destroyed_wall_positions.add(living_wall_position)


        # Figuring out which players were shot, removing these bullets:
        new_player_positions_that_were_shot = set()
        for new_player_position, old_player_position in new_player_position_to_old.items():
            if wip_bullets.get(new_player_position, None):
                # Common shooting case.
                del wip_bullets[new_player_position]
                new_player_positions_that_were_shot.add(new_player_position)
            elif translation := new_player_position - old_player_position: # Player moved
                oncoming_bullet_direction = - Step(*translation)
                oncoming_bullets = {bullet for bullet in wip_bullets.get(old_player_position, ()) if
                                    bullet.direction == oncoming_bullet_direction}
                if oncoming_bullets:
                    # Less-common shooting case: The player walked towards an oncoming bullet,
                    # switching places with it.
                    (oncoming_bullet,) = oncoming_bullets
                    wip_bullets[old_player_position].remove(oncoming_bullet)
                    new_player_positions_that_were_shot.add(new_player_position)


        bullets = ImmutableDict({
            position: frozenset(bullets) for position, bullets in wip_bullets.items() if bullets
        })

        #                                                                                          #
        ### Finished figuring out bullets. #########################################################
        ############################################################################################

        ############################################################################################
        ### Figuring out food: #####################################################################
        #                                                                                          #
        new_player_positions_that_ate_food = set()
        wip_food_positions = set(self.food_positions)
        random_positions_firehose = utils.iterate_deduplicated(
            State.iterate_random_positions(board_size=self.board_size),
            seen=(set(itertools.chain(new_player_position_to_old, self.position_to_letter)) |
                  self.food_positions)
        )
        for new_player_position in new_player_position_to_old:
            if new_player_position in self.food_positions:
                wip_food_positions.remove(new_player_position)
                wip_food_positions.add(next(random_positions_firehose))
                new_player_positions_that_ate_food.add(new_player_position)

        assert len(wip_food_positions) == len(self.food_positions)

        #                                                                                          #
        ### Finished figuring out food. ############################################################
        ############################################################################################

        player_id_to_observation = {}

        for new_player_position, old_player_position in new_player_position_to_old.items():
            letter = self.position_to_letter[old_player_position]
            old_observation: Observation = self[letter]

            reward = (
                SHOT_REWARD if new_player_position in new_player_positions_that_were_shot else
                COLLISION_REWARD if new_player_position in collided_player_positions else
                FOOD_REWARD if new_player_position in new_player_positions_that_ate_food else
                NOTHING_REWARD
            )

            player_id_to_observation[letter] = Observation(
                state=None,
                position=new_player_position,
                score=old_observation.score + reward,
                reward=reward,
                letter=letter,
                last_action=activity[letter]
            )


        state = State(
            board_size=self.board_size,
            player_id_to_observation=player_id_to_observation,
            food_positions=frozenset(wip_food_positions), bullets=bullets,
            allow_shooting=self.allow_shooting, allow_walling=self.allow_walling,
            living_wall_positions=frozenset(wip_living_wall_positions),
            destroyed_wall_positions=frozenset(wip_destroyed_wall_positions),
        )

        for observation in player_id_to_observation.values():
            observation.state = state

        payoff = gamey.Payoff(
            player_id_to_reward={letter: observation.reward for letter, observation in
                                 player_id_to_observation.items()}
        )
        return (payoff, state)

    @property
    def ascii(self) -> None:
        '''A nice ascii representation of all the objects in the state.'''
        string_io = io.StringIO()
        for position in Position.iterate_all(self):
            if position.x == 0 and position.y != 0:
                string_io.write('|\n')
            if position in self.position_to_letter:
                letter = self.position_to_letter[position]
                observation = self[letter]
                string_io.write(letter.lower() if observation.reward < NOTHING_REWARD
                                else observation.letter)
            elif (bullets := self.bullets.get(position, None)):
                if len(bullets) >= 2:
                    string_io.write('Ӿ')
                else:
                    string_io.write(more_itertools.one(bullets).direction.ascii)


            elif position in self.food_positions: # But no bullets
                string_io.write('·')
            else:
                string_io.write(' ')

        string_io.write('|\n')
        string_io.write('‾' * self.board_size)
        string_io.write('\n')

        return string_io.getvalue()

    def p(self) -> None:
        print(self.ascii)




    @property
    def average_reward(self) -> numbers.Real:
        return statistics.mean(observation.reward for observation in self.values())


class Policy(gamey.ModelFreeLearningPolicy):
    default_block_size = 750_000
    observation_type = Observation
    action_type: Type[Action] = SavvyField(lambda: Action)

    @classmethod
    def create(cls, *, board_size: Optional[int] = None, **kwargs) -> None:
        if 'observation_neural_dtype' not in kwargs:
            kwargs['observation_neural_dtype'] = Observation.get_neural_dtype_for_board_size(
                board_size if board_size is not None else DEFAULT_BOARD_SIZE
            )
        return gamey.ModelFreeLearningPolicy.create.__func__(cls, **kwargs)


    @staticmethod
    def create_model(observation_neural_dtype: np.dtype, action_n_neurons: int) -> 'keras.Model':
        from tensorflow import keras

        grid_input = keras.Input(
            shape=observation_neural_dtype['grid_input'].shape,
            name='grid_input'
        )
        grid_0 = keras.layers.Conv2D(
            16, 2, activation='relu',
            kernel_initializer='orthogonal'
        )(grid_input)
        _ = keras.layers.Dropout(rate=0.1)(grid_0)
        _ = keras.layers.Conv2D(
            16, 2, activation='relu',
            kernel_initializer='orthogonal'
        )(_)
        _ = keras.layers.Dropout(rate=0.1)(_)
        _ = keras.layers.Conv2D(
            16, 2, activation='relu',
            kernel_initializer='orthogonal'
        )(_)
        _ = keras.layers.Flatten()(_)
        grid_output = keras.layers.Dropout(rate=0.1)(_)

        sequential_input = keras.Input(
            shape=observation_neural_dtype['sequential_input'].shape,
            name='sequential_input'
        )

        concatenate_layer = keras.layers.concatenate([grid_output, sequential_input])

        _ = keras.layers.Dense(
            128, activation='relu',
        )(concatenate_layer)
        _ = keras.layers.Dropout(rate=0.1)(_)
        _ = keras.layers.Dense(
            128, activation='relu',
        )(_)
        _ = keras.layers.Dropout(rate=0.1)(_)
        output = keras.layers.Dense(
            more_itertools.one(action.to_neural().shape),
        )(_)
        model = keras.Model([grid_input, sequential_input], output)
        model.compile(optimizer='rmsprop', loss='mse', metrics=['accuracy'])

        return model



class Game(gamey.Game):
    def write_to_jam_parchment(self, jam_parchment: jamming.JamParchment, *,
                               max_length: Optional[int] = None) -> Iterator[State]:
        from .animating import animate

        state_iterator, state_iterator_copy = itertools.tee(
                                                  more_itertools.islice_extended(self)[:max_length])
        transition_iterator = animate(state_iterator_copy)
        while True:
            limited_iterator = more_itertools.islice_extended(
                                         zip(state_iterator, transition_iterator, strict=True))[:10]
            transition_chunk = []
            for state, transition in limited_iterator:
                transition_chunk.append(transition)
                yield state
            self.save()
            if transition_chunk:
                jam_parchment.extend_jams(transition_chunk)
            else:
                # Game ended, either naturally or reached `max_length`.
                return





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import functools
import warnings
import re
import json
import urllib.parse
import threading
import pathlib
import requests
import logging
from typing import (Optional, Tuple, Union, Container, Hashable, Iterator,
                    Iterable, Any, Dict, FrozenSet, Callable, Type, Sequence)

import flask

from marley import constants
from marley.worlds.grid_royale.core import get_games_folder
from marley.jamswank.jamming import server as jamming_server
from marley.jamswank.jamming.jam_file_database import JamFileDatabase


DEFAULT_HOST = '127.0.0.1'
DEFAULT_PORT = 22390
STATIC_FOLDER = pathlib.Path(__file__).parent / 'static'

assert STATIC_FOLDER.exists()
assert STATIC_FOLDER.is_dir()

def make_url_from_host_port(host: str, port: Union[str, int]):
    return urllib.parse.urlunsplit(('http', f'{host}:{port}', '', '', ''))

DEFAULT_URL = make_url_from_host_port(DEFAULT_HOST, DEFAULT_PORT)


app = flask.Flask('ravenna', static_url_path='/static', static_folder=str(STATIC_FOLDER))
jamming_server.add_to_flask_app(constants.jam_file_database, app)

@app.route('/')
def index_redirect():
    return flask.redirect('/static/index.html')


@app.route('/shutdown', methods=('POST',))
def shutdown() -> str:
    shutdown_function = flask.request.environ.get('werkzeug.server.shutdown')
    if shutdown_function is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    shutdown_function()
    return 'Shut down flask server.'

@app.route('/games')
def games() -> str:
    return json.dumps([path.name for path in get_games_folder().iterdir()
                       if path.is_dir() and any(path.iterdir())])

@app.route('/games/<game>/<chunk>.json')
def game_chunk(game: str, chunk: str) -> str:
    assert re.fullmatch('^[0-9]+$', chunk)
    try:
        return (get_games_folder() / game / f'{chunk}.json').read_text()
    except FileNotFoundError:
        raise flask.abort(404)

def _shut_up_flask():
    logger = logging.getLogger('werkzeug')
    logger.setLevel(logging.ERROR)
    app.logger.disabled = logger.disabled = True
    flask.cli.show_server_banner = lambda *args, **kwargs: None
    warnings.filterwarnings(
        'ignore',
        message="The 'environ['werkzeug.server.shutdown']' function is deprecated"
    )


class ServerThread(threading.Thread):
    def __init__(self, host: str = DEFAULT_HOST, port: Union[int, str] = DEFAULT_PORT, *,
                 quiet: bool = False) -> None:
        threading.Thread.__init__(self, daemon=True)
        self.host = host
        self.port = int(port)
        self.quiet = quiet

    def run(self) -> None:
        if self.quiet:
            _shut_up_flask()
        app.run(host=self.host, port=self.port)

    def shutdown(self) -> None:
        response = requests.post(urllib.parse.urljoin(DEFAULT_URL, 'shutdown'))
        response.raise_for_status()
        self.join()

    def __enter__(self) -> ServerThread:
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.shutdown()

    @property
    def url(self):
        host = '127.0.0.1' if self.host == '0.0.0.0' else self.host
        return make_url_from_host_port(host, self.port)





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import pathlib
import logging
import functools
import json

from marley.jamswank.jamming.jam_file_database import JamFileDatabase
from .marley_swank_database import MarleySwankDatabase

@functools.cache
def read_config() -> dict:
    try:
        content = config_path.read_text()
    except FileNotFoundError:
        return {}
    return json.loads(content)



marley_folder: pathlib.Path = pathlib.Path.home() / '.marley'
config_path: pathlib.Path = marley_folder / 'config.json'
logs_folder: pathlib.Path = marley_folder / 'logs'
config = read_config() # todo: Shouldn't be on import

if 'jam_database_path' in config:
    jam_database_folder = pathlib.Path(config['jam_database_path'])
    assert jam_database_folder.exists()
else:
    jam_database_folder: pathlib.Path = marley_folder / 'jam_database'


jam_file_database = JamFileDatabase(jam_database_folder)
swank_database = MarleySwankDatabase(jam_file_database)




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import re
import os
import sys
import io
import abc
import itertools
import bisect
import pathlib
import json
import collections.abc
from typing import Union, Iterable, Any, Mapping, TypeVar, Optional, Tuple, Sequence, Sized
import datetime as datetime_module
import string
import random

from . import utils

CORE_LENGTH = 12

Jam = TypeVar('Jam', bound=dict)

jam_id_pattern = re.compile(
    r'^([0-9a-z]{' + str(CORE_LENGTH) + r'})\.([0-9]{7})$'
)
datetime_format = '%Y%m%d-%H%M%S-%f'
digits_and_lowercase_ascii = string.digits + string.ascii_lowercase

class JamException(Exception):
    pass

class JamCorruption(JamException):
    pass

class JamTooLong(JamException):
    pass


class DateTimeSupportingJsonEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime_module.datetime):
            return o.isoformat()

        return json.JSONEncoder.default(self, o)


def fixed_size_json(jam: Jam, size: int) -> bytes:
    x: bytes = json.dumps(jam, cls=DateTimeSupportingJsonEncoder).encode('utf-8')
    if len(x) > size:
        raise JamTooLong
    else:
        return x.ljust(size, b' ')

def cute_json_load(text: str) -> Any:
    if text.startswith('\0'):
        return None
    else:
        return json.loads(text)



class JamId:
    def __init__(self, jam_id_or_name: Union[JamId, str]) -> None:
        self.text = jam_id_or_name if isinstance(jam_id_or_name, str) else str(jam_id_or_name)
        assert '/' not in self.text
        assert '\\' not in self.text
        match = jam_id_pattern.fullmatch(self.text)
        assert match is not None
        self.block_size = int(match.groups()[-1])

    @property
    def datetime(self):
        try:
            return self._datetime
        except AttributeError:
            self._datetime = datetime_module.datetime.strptime(self.text, datetime_format)
            return self._datetime

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f'{type(self).__name__}({repr(self.text)})'

    @staticmethod
    def parse(jam_id_or_str: Union[str, JamId]) -> Tuple[JamId, str]:
        if isinstance(jam_id_or_str, str):
            jam_id_name = jam_id_or_str
            jam_id = JamId(jam_id_name)
        else:
            assert isinstance(jam_id_or_str, JamId)
            jam_id_name = str(jam_id_or_str)
            jam_id = jam_id_or_str
        return (jam_id, jam_id_name)

    @staticmethod
    def cast(jam_id_or_str: Union[str, JamId]) -> JamId:
        if isinstance(jam_id_or_str, str):
            return JamId(jam_id_or_str)
        else:
            assert isinstance(jam_id_or_str, JamId)
            return jam_id_or_str

    @staticmethod
    def create(*, block_size: int = 1_000, zero: bool = False) -> JamId:
        assert isinstance(block_size, int)
        assert block_size >= 1
        if zero:
            content = '0' * CORE_LENGTH
        else:
            content = ''.join(random.choices(digits_and_lowercase_ascii, k=CORE_LENGTH))
        return JamId(
            f'{content}.{block_size:07d}'
        )

    def _reduce(self) -> tuple:
        return (type(self), self.text)


    def __eq__(self, other: JamId) -> bool:
        return (type(self) is type(other)) and (self._reduce() == other._reduce())

    def __hash__(self) -> int:
        return hash(self._reduce())





class BaseJamReducable:
    @abc.abstractmethod
    def _reduce(self) -> Tuple[Any]:
        raise NotImplementedError

    def __eq__(self, other: Any) -> bool:
        return (type(self) is type(other)) and (self._reduce() == other._reduce())

    def __repr__(self) -> str:
        return f'{type(self).__name__}({", ".join(map(repr, self._reduce()))})'


class BaseJamDatabase(abc.ABC, Iterable, BaseJamReducable):
    def __init__(self, path: Union[pathlib.Path, str]) -> None:
        self.path = pathlib.Path(path)

    def __getitem__(self, name_or_type: Union[str, type]) -> JamKind:
        if isinstance(name_or_type, str):
            name = name_or_type
        else:
            name = utils.type_to_name(name_or_type)
        return JamKind(self, name)

    def __iter__(self) -> Iterable[JamKind]:
        yield from self._get_jam_kinds()

    def __len__(self):
        return len(self._get_jam_kinds())

    @abc.abstractmethod
    def _get_jam_kinds(self) -> Sequence[JamKind]:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_jam_kind_length(self, jam_kind_name: str) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_jam_parchments(self, jam_kind_name: str) -> Sequence[JamParchment]:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_jam_parchment_length(self, jam_kind_name: str, jam_id: Union[str, JamId]) -> int:
        raise NotImplementedError


    @abc.abstractmethod
    def _read_texts(self, jam_kind_name: str, jam_id: Union[str, JamId],
                    start: int, end: Optional[int]) -> Tuple[str]:
        raise NotImplementedError


class BaseWritableJamDatabase(BaseJamDatabase):
    @abc.abstractmethod
    def _write_blobs(self, jam_kind_name: str, jam_id: Union[str, JamId],
                     texts: Iterable[bytes], *, start: int = 0, append: bool = False) -> None:
        raise NotImplementedError


class JamKind(BaseJamReducable):
    def __init__(self, jam_database: BaseJamDatabase, name: str) -> None:
        assert re.fullmatch('[a-zA-Z0-9_.]+', name)
        self.jam_database = jam_database
        self.name = name

    def _reduce(self) -> Tuple[Any]:
        return (self.jam_database, self.name)

    def __len__(self) -> int:
        return self.jam_database._get_jam_kind_length(self.name)

    def __iter__(self) -> Iterable[JamParchment]:
        yield from self.jam_database._get_jam_parchments(self.name)

    def __getitem__(self, jam_id: Union[JamId, str]) -> JamParchment:
        return JamParchment(self, JamId(jam_id))


class JamParchment(collections.abc.Sequence, BaseJamReducable):
    def __init__(self, jam_kind: JamKind, jam_id: JamId) -> None:
        self.jam_kind = jam_kind
        self.jam_id = jam_id
        self.block_size = jam_id.block_size
        self.jam_database = self.jam_kind.jam_database

    def _reduce(self) -> Tuple[Any]:
        return (self.jam_database, self.jam_kind, self.jam_id)

    def __iter__(self) -> Iterable[JamItem]:
        for i in itertools.count():
            yield JamItem(self, i)

    def __len__(self) -> int:
        return self.jam_database._get_jam_parchment_length(self.jam_kind.name, self.jam_id)

    def __getitem__(self, i: int) -> JamItem:
        assert isinstance(i, int) # Slices not supported
        return JamItem(self, i)

    def extend_blobs(self, blobs: Iterable[bytes]) -> None:
        assert isinstance(self.jam_database, BaseWritableJamDatabase)
        self.jam_database._write_blobs(self.jam_kind.name, self.jam_id, blobs, append=True)

    def extend_jams(self, jams: Iterable[Jam]) -> None:
        self.extend_blobs(fixed_size_json(jam, self.block_size) for jam in jams)

    def read_texts(self, start: int, end: Optional[int]) -> Tuple[str, ...]:
        return self.jam_database._read_texts(self.jam_kind.name, self.jam_id, start, end)

    def delete(self):
        self._get_path().unlink()

    def _get_path(self) -> os.PathLike:
        # Todo: This method makes me want to refactor jamming to have the path on each object.
        from .jam_file_database import JamFileDatabase
        assert isinstance(self.jam_database, JamFileDatabase)
        return self.jam_database.path / self.jam_kind.name / str(self.jam_id)

    def _get_lock_path(self) -> os.PathLike:
        path = self._get_path()
        return type(path)(str(path) + '.lock')





class JamItem(BaseJamReducable):
    def __init__(self, jam_parchment: JamParchment, jam_index: int) -> None:
        self.jam_parchment = jam_parchment
        self.jam_database = self.jam_parchment.jam_database
        self.jam_index = jam_index

    def _reduce(self) -> Tuple[Any]:
        return (self.jam_database, self.jam_parchment, self.jam_index)

    def read_text(self) -> bytes:
        end = (self.jam_index + 1) if self.jam_index != -1 else None
        texts = self.jam_parchment.read_texts(self.jam_index, end)

        if texts:
            (text,) = texts
            return text
        else:
            raise IndexError(self.jam_index)

    def write_blob(self, blob: bytes) -> None:
        assert isinstance(self.jam_database, BaseWritableJamDatabase)
        self.jam_database._write_blobs(self.jam_parchment.jam_kind.name,
                                       self.jam_parchment.jam_id, (blob,), start=self.jam_index)

    def read_jam(self) -> Jam:
        return cute_json_load(self.read_text())

    def write_jam(self, jam: Jam) -> None:
        self.write_blob(fixed_size_json(jam, self.jam_parchment.block_size))




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

'''This module defines classes for different kinds of vectors, and common actions on them.'''

from __future__ import annotations

import dataclasses
import itertools
import operator
import functools
import math
from typing import (Optional, Tuple, Union, Container, Hashable, Iterator,
                    Iterable, Any, Dict, FrozenSet)
import numpy as np

from . import core


@dataclasses.dataclass(frozen=True)
class Vector:
    '''A vector in 2 dimensions, e.g. Vector(-2, 3)'''

    x: int
    y: int

    __iter__ = lambda self: iter((self.x, self.y))

    def __matmul__(self, other: Vector) -> int:
        '''Get Levenshtein distance between two vectors'''
        return sum(map(abs, map(operator.sub, self, other)))

    def __neg__(self) -> Vector:
        return type(self)(x=-self.x, y=-self.y)

    __bool__ = lambda self: any(self)

    def rotate_in_board(self, n_right_angles: int, /, board_size: int) -> Vector:
        x, y = self
        for _ in range(n_right_angles % 4):
            (x, y) = (board_size - 1 - y, x)
        return type(self)(x, y)

    def iterate_rotations_in_board(self, /, board_size: int) -> Iterator[Vector]:
        x, y = self
        yield self
        for _ in range(3):
            (x, y) = (board_size - 1 - y, x)
            yield type(self)(x, y)

    def __hash__(self):
        return hash((type(self), *self))

    def __eq__(self, other: Any):
        return (type(self) is type(other)) and (tuple(self) == tuple(other))




class Position(Vector):
    @staticmethod
    def iterate_all(state_or_board_size: Union[core.Grid, int], /):
        '''Iterate over all positions in a board of the given size.'''

        board_size: int = (state_or_board_size if isinstance(state_or_board_size, int)
                           else state_or_board_size.board_size)
        for y, x in itertools.product(range(board_size), repeat=2):
            yield Position(x, y)

    def __sub__(self, other: Union[Position, Translation]):
        if isinstance(other, Position):
            result_type = Translation
        else:
            assert isinstance(other, Translation)
            result_type = Position

        return result_type(self.x - other.x, self.y - other.y)


    def __add__(self, translation: Translation):
        assert isinstance(translation, Translation)
        return Position(self.x + translation.x,
                        self.y + translation.y)


    def in_square(self, square_size: int) -> bool:
        '''Check whether the position is in bounds of a square of the given size.'''
        return ((0 <= self.x <= square_size - 1) and
                (0 <= self.y <= square_size - 1))

    def walk(self, translation: Translation,
             board_size: Optional[int] = None) -> Iterator[Position]:
        position = self
        if board_size is None:
            while True:
                yield position
                position += translation
        else:
            while 0 <= min(position) <= max(position) <= board_size - 1:
                yield position
                position += translation


    @functools.cache
    def field_of_view(self, vicinity: Vicinity, board_size: int) -> Tuple[FrozenSet[Position]]:
        result = []
        count_from_one = itertools.count(start=1)
        if vicinity._is_step_like():
            for i in count_from_one:
                positions = frozenset(
                    position for j in range(-i, i + 1) if (position := (self +
                    i * vicinity + j * Step(*vicinity).perpendiculars[0])).in_square(board_size)
                )
                if not positions:
                    return tuple(result)
                result.append(positions)
        else:
            for i in count_from_one:
                positions = frozenset(
                    position for j in range(i + 1) if (position := (
                    self + Translation(j * vicinity.x, (i - j) * vicinity.y))).in_square(board_size)
                )
                if not positions:
                    return tuple(result)
                result.append(positions)


    def horizontal_line_in_board(self, board_size: int) -> Tuple[Position]:
        return tuple(Position(i, self.y) for i in range(board_size))

    def vertical_line_in_board(self, board_size: int) -> Tuple[Position]:
        return tuple(Position(self.x, i) for i in range(board_size))








PositionLike = Union[Position, Tuple[int, int]]

class Translation(Vector):
    '''A Translation (i.e. movement) in 2-dimensional space.'''
    def __mul__(self, number: int) -> Translation:
        assert isinstance(number, int)
        return type(self)(x=(self.x * number), y=(self.y * number))

    def __rmul__(self, number: int) -> Translation:
        return self * number

    def _is_step_like(self):
        return tuple(self).count(0) == 1


class Vicinity(Translation):
    '''A movement like a king in chess, i.e. 8 possible positions around the origin.'''
    all_vicinities: Tuple[Vicinity]
    def __init__(self, x: int, y: int):
        assert {0} != set((x, y)) <= {-1, 0, 1}
        super().__init__(x=x, y=y)


    def __mul__(self, number: int) -> Translation:
        assert isinstance(number, int)
        return Translation(x=(self.x * number), y=(self.y * number))

Vicinity.all_vicinities = tuple(
    itertools.starmap(
        Vicinity,
        filter(
            any,
            itertools.product((-1, 0, 1), repeat=2)
        )
    )
)


class Step(Vicinity):

    _all_ascii = '↑→↓←'

    def __init__(self, x: int, y: int):
        assert frozenset((x, y)) in {frozenset((0, 1)),
                                     frozenset((0, -1))}
        super().__init__(x=x, y=y)

    @property
    def index(self):
        try:
            return self._index
        except AttributeError:
            self._index = tuple(Step.all_steps).index(self)
            return self._index

    @property
    def name(self) -> str:
        try:
            return self._name
        except AttributeError:
            self._name = ('up', 'right', 'down', 'left')[self.index]
            return self._name

    @property
    def ascii(self) -> str:
        try:
            return self._ascii
        except AttributeError:
            self._ascii = Step._all_ascii[tuple(Step.all_steps).index(self)]
            return self._ascii

    def is_general_direction(self, source: Position, target: Position) -> bool:
        translation: Translation = target - source
        if self.x:
            return ((np.sign(translation.x) == self.x) and
                    abs(translation.x) >= abs(translation.y))
        else:
            assert self.y
            return ((np.sign(translation.y) == self.y) and
                    abs(translation.y) >= abs(translation.x))


    @property
    def perpendiculars(self) -> Tuple[Step]:
        try:
            return self._perpendiculars
        except AttributeError:
            self._perpendiculars = tuple(
                sorted(
                    ((first := type(self)(*reversed(tuple(self)))), -first),
                    key=tuple
                )
            )
            return self._perpendiculars

    @property
    def angle_from_top(self):
        return math.tau * (self.index / 4)


(Step.up, Step.right, Step.down, Step.left) = Step.all_steps = (
    Step(0, -1), Step(1, 0), Step(0, 1), Step(-1, 0)
)



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

'''
Gamey is a cute reinforcement learning framework.

I wrote Gamey so I would have a clear separation between generic reinforcement learning architecture
and the specific game of GridRoyale. Gamey defineds states, observations, actions, policies,
cultures. It defines these as base classes that should be overridden.

See the `sample_games/blackjack.py` module for an example of using Gamey on Blackjack.
'''

from __future__ import annotations

from typing import Optional

from .base import Observation, Action, ActionEnum, Story, PlayerId
from .policing import Policy, RandomPolicy, SoloEpisodicPolicy, QPolicy
from .aggregating import State, SoloState, Payoff, Activity, Culture
from .gaming import Game, Narrative
from .model_free import ModelFreeLearningPolicy
from .model_based import ModelBasedEpisodicLearningPolicy
from . import exceptions
from marley.gamey.gamey_swanking import GameySwankDatabase, GameySwank


global_gamey_swank_database: GameySwankDatabase = GameySwankDatabase.create_ethereal()



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import math
import inspect
import re
import abc
import random
import itertools
import collections.abc
import statistics
import concurrent.futures
import enum
import functools
import numbers
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable)
import dataclasses

import more_itertools
import numpy as np

from marley import gamey
from .policing import Policy
from .base import Observation, Action
from . import utils


@dataclasses.dataclass(order=True, frozen=True)
class ActionObservation(utils.NiceDataclass):
    action: Optional[Action]
    observation: Observation



class ModelBasedEpisodicLearningPolicy(Policy):
    '''
    Model-based episodic learning policy.

    This policy assumes we're playing full episodes to the end, and there is no reward
    discounting.
    '''
    def __init__(self, curiosity: numbers.Real = 2) -> None:
        self.reward_map = RewardMap()
        self.curiosity = curiosity
        self.action_observation_chains_lists = collections.defaultdict(list)

    def get_next_action(self, observation: Observation) -> Action:
        return max(
            observation.legal_actions,
            key=lambda action: self.reward_map.get_ucb(
                observation, action, curiosity=self.curiosity
            )
        )

    def train(self, narratives: Sequence[gamey.Narrative]) -> ModelBasedEpisodicLearningPolicy:
        raise NotImplementedError

        action_observation_chains = self.action_observation_chains_lists[observation]
        try:
            action_observation_chain = action_observation_chains.pop()
        except IndexError:
            action_observation_chain = [ActionObservation(None, observation)]

        action_observation_chain.append(ActionObservation(action, next_observation))

        if next_observation.state.is_end:
            total_reward = 0
            for new_action_observation, old_action_observation in \
                                   utils.iterate_windowed_pairs(reversed(action_observation_chain)):
                total_reward += new_action_observation.observation.reward
                self.reward_map.add_sample(old_action_observation.observation,
                                           new_action_observation.action, total_reward)
        else:
            self.action_observation_chains_lists[next_observation].append(action_observation_chain)



def _zero_maker():
    return 0


class RewardMap(collections.abc.Mapping):
    def __init__(self) -> None:
        self._reward_values = collections.defaultdict(_zero_maker)
        self._n_samples = collections.defaultdict(_zero_maker)
        self.n_total_samples = 0

    __len__ = lambda self: len(self._reward_values)
    __iter__ = lambda self: iter(self._reward_values)

    def __getitem__(self, observation_and_action: Iterable) -> numbers.Real:
        return self._reward_values[self._to_key(*observation_and_action)]

    def _to_key(self, observation: Observation, action: Action) -> Tuple[Observation, Action]:
        assert isinstance(observation, Observation)
        assert isinstance(action, Action)
        return (observation, action)



    def add_sample(self, observation: Observation, action: Action, reward: numbers.Real) -> None:
        key = self._to_key(observation, action)
        self._reward_values[key] = (
            self._reward_values[key] *
            (self._n_samples[key] / (self._n_samples[key] + 1)) +
            reward * (1 / (self._n_samples[key] + 1))
        )
        self._n_samples[key] += 1
        self.n_total_samples += 1

    def get_ucb(self, observation: Observation, action: Action,
                curiosity: numbers.Real) -> numbers.Real:
        key = self._to_key(observation, action)
        return self._reward_values[key] + curiosity * math.sqrt(
            utils.cute_div(
                math.log(self.n_total_samples + 2),
                self._n_samples[key]
            )
        )





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import datetime as datetime_module
from typing import Iterable, Mapping, Union, Type, Optional, Any
import itertools
import collections.abc

import more_itertools

from ...jamming import BaseJamDatabase, Jam, JamId, JamItem, JamKind, JamParchment
from ..core import Swank, SwankDatabase
from .base_field import BaseField
from .. import utils
from ..exceptions import EmptyJam


class CantExtendBoundedParchmentRef(Exception):
    pass



class ParchmentRef(collections.abc.MutableSequence):
    def __init__(self, swank_database: SwankDatabase, swank_type: Optional[Type[Swank]] = None,
                 jam_id: Union[JamId, str, None] = None, start_index: int = 0,
                 end_index: Optional[int] = None) -> None:
        assert (swank_type, jam_id).count(None) in (0, 2)
        self.swank_database = swank_database
        self.swank_type = swank_type
        self.jam_id = JamId.parse(jam_id)[0] if jam_id is not None else None
        self.start_index = start_index
        self.end_index = end_index
        self.cached_swanks = {}
        self.cached_modified_indices = set()
        self.cached_length = None

    def migrate(self, jam_id: Union[JamId, str, None] = None, start_index: int = 0,
                end_index: Optional[int] = None, *, delete_old_jam_parchment: bool = True) -> None:
        assert self.is_specified
        old_jam_parchment = self.jam_parchment

        for i, swank in enumerate(self):
            jam_index = i + start_index
            if end_index is not None:
                assert jam_index < end_index
            swank.jam_id = jam_id
            swank.jam_index = jam_index
            swank.save()
            self.cached_modified_indices.discard(i)
        assert not self.cached_modified_indices

        self.jam_id = jam_id
        self.start_index = start_index
        self.end_index = end_index

        self.save()

        if delete_old_jam_parchment and old_jam_parchment != self.jam_parchment:
            old_jam_parchment.delete()


    def save(self) -> None:
        while self.cached_modified_indices:
            i = self.cached_modified_indices.pop()
            self.cached_swanks[i].save()
        self.cached_swanks.clear()

    @property
    def is_specified(self) -> bool:
        count = (self.swank_type, self.jam_id).count(None)
        assert count in (0, 2)
        return not count

    @property
    def is_bounded(self) -> bool:
        result = (self.end_index is not None)
        return result

    def specify(self, swank_or_type: Union[Swank, Type[Swank]]) -> None:
        swank_type = type(swank_or_type) if isinstance(swank_or_type, Swank) else swank_or_type
        self.swank_type = swank_type
        self.jam_id = JamId.create(block_size=swank_type.default_block_size)


    @property
    def jam_kind_name(self) -> Optional[str]:
        if self.is_specified:
            return utils.type_to_name(self.swank_type)
        else:
            return None

    @property
    def jam_parchment(self) -> Optional[JamParchment]:
        if self.is_specified:
            return self.swank_database.jam_database[self.swank_type][self.jam_id]
        else:
            return None

    def __getitem__(self, i: Union[int, slice]) -> Swank:
        if isinstance(i, slice):
            raise NotImplementedError
        if not self.is_specified:
            raise IndexError(i)
        if i < 0:
            i += len(self)
            if not 0 <= i < len(self):
                raise IndexError(i)
        if self.is_bounded and i >= (self.end_index - self.start_index):
            raise IndexError(i)
        try:
            return self.cached_swanks[i]
        except KeyError:
            try:
                self.cached_swanks[i] = self.swank_database.load_swank(
                    self.swank_type, self.jam_id, i + self.start_index
                )
            except EmptyJam as empty_jam:
                raise IndexError(i) from empty_jam
            else:
                return self.cached_swanks[i]

    def __setitem__(self, i: Union[int, slice], swank: Swank) -> None:
        if isinstance(i, slice):
            raise NotImplementedError
        if i < 0:
            raise NotImplementedError
        if not self.is_specified:
            self.specify(type(swank))

        if self.is_bounded and i >= (self.end_index - self.start_index):
            raise IndexError(i)

        swank.jam_id = self.jam_id
        swank.jam_index = i + self.start_index
        self.cached_swanks[i] = swank
        self.cached_modified_indices.add(i)
        if (self.cached_length is None) or i >= self.cached_length:
            self.cached_length = i + 1

    def __len__(self) -> int:
        if self.cached_length is None:
            if self.is_bounded:
                self.cached_length = self.end_index - self.start_index
            else:
                self.cached_length = ((len(self.jam_parchment) - self.start_index)
                                      if self.is_specified else 0)
        return self.cached_length


    def insert(self, i: int, swank: Swank) -> None:
        raise NotImplementedError

    def __delitem__(self, i: int) -> None:
        raise NotImplementedError

    def append(self, swank: Swank) -> None:
        self.extend((swank,))

    def extend(self, swanks: Iterable[Swank]) -> None:
        if self.is_bounded:
            raise CantExtendBoundedParchmentRef
        first_run = True
        for i, swank in enumerate(swanks, start=len(self)):
            if first_run and (not self.is_specified):
                self.swank_type = type(swank)
                self.jam_id = JamId.create(block_size=type(swank).default_block_size)

            first_run = False

            assert isinstance(swank, self.swank_type)
            swank.jam_id = self.jam_id
            swank.jam_index = i + self.start_index

            self.cached_swanks[i] = swank
            self.cached_modified_indices.add(i)
            self.cached_length += 1

    def clear(self):
        raise NotImplementedError

    def reverse(self):
        raise NotImplementedError

    def pop(self, index=-1):
        raise NotImplementedError

    def remove(self, value):
        raise NotImplementedError

    def __repr__(self) -> str:
        items = (
            str(self.swank_database),
            utils.type_to_name(self.swank_type) if (self.swank_type is not None) else 'None',
            repr(str(self.jam_id)),
            *((str(self.start_index),) if self.start_index != 0 else ()),
            *((str(self.end_index),) if self.start_index is not None else ()),
        )
        return f'{type(self).__name__}({", ".join(items)})'

    def has_index(self, i: int) -> bool:
        try:
            self[i]
        except IndexError:
            return False
        else:
            return True




class ParchmentField(BaseField):

    field_type_name = 'parchment'

    def from_jam(self, jam: Jam, swank_database: SwankDatabase, *,
                 swank_cache: dict) -> ParchmentRef:
        if jam is None:
            return self.get_default_value(swank_database)
        else:
            jam_kind_name, jam_id_name, start_index, end_index = jam
            if jam_kind_name is None:
                assert jam_id_name is None
                assert start_index == 0
                assert end_index is None
                return ParchmentRef(swank_database)
            else:
                return ParchmentRef(swank_database, utils.name_to_type(jam_kind_name),
                                    JamId(jam_id_name), start_index, end_index)

    def to_jam(self, value: Optional[ParchmentRef], swank_database: SwankDatabase) -> Jam:
        return (
            value.jam_kind_name,
            (str(value.jam_id) if value.jam_id else None),
            value.start_index,
            value.end_index,
        )

    def get_default_value(self, swank_database: SwankDatabase) -> ParchmentRef:
        return ParchmentRef(swank_database)

    def __set__(self, swank: Swank, other_swanks: Union[Iterable[Swank], ParchmentRef]) -> None:
        if isinstance(other_swanks, ParchmentRef):
            swank._Swank__field_values[self.name] = other_swanks
        else:
            other_swanks = tuple(other_swanks)
            if not other_swanks:
                if self.name not in swank._Swank__field_values:
                    swank._Swank__field_values[self.name] = ParchmentRef(swank.swank_database)
                elif swank._Swank__field_values[self.name]:
                    raise NotImplementedError
                else:
                    pass # Assigned empty to empty, we'll just no-op.
                return
            other_swank_type = type(other_swanks[0])
            assert all((type(other_swank) is other_swank_type) for other_swank in other_swanks)
            new_parchment_ref = ParchmentRef(
                swank.swank_database, other_swank_type,
                JamId.create(block_size=other_swank_type.default_block_size)
            )
            swank._Swank__field_values[self.name] = new_parchment_ref
            new_parchment_ref.extend(other_swanks)




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import json
import math
import sys
import re
import time
import itertools
import numbers
import collections.abc
import urllib.parse
from typing import Union, Optional, Iterator, Iterable, TypeVar, Callable, Tuple, TypeVar, Sequence
import random

from browser import document, html, ajax, timer, window

from .utils import cool_ajax, add_parameters_to_url, sliced


jam_id_pattern = re.compile(
    r'^([0-9]{8})-([0-9]{6})-([0-9]{6})-([0-9a-z]{30})\.([0-9]{12})$'
)
jam_kind_pattern = re.compile('[a-z_]+')


class JellyDatabase:
    def __init__(self, url: str) -> None:
        assert url.endswith('/')
        self.url = url


    def _make_request(self, path: str, callback: Callable, *, binary: bool = False,
                      disable_cache: bool = True) -> None:
        url = f'{self.url}{path}'
        if disable_cache:
            url = add_parameters_to_url(url, {'_': random.randint(0, 10**8)})
        def inner(ajax):
            try:
                callback(ajax)
            except:
                import sys
                sys.excepthook(*sys.exc_info())
                raise

        ajax.get(url, mode=('binary' if binary else 'text'), oncomplete=inner)


    def get_jam_parchment_names(self, jam_kind_name: str,
                                callback: Callable[[Sequence[str]], None]) -> None:
        assert jam_kind_pattern.fullmatch(jam_kind_name)
        def inner(request: ajax.ajax) -> None:
            jam_parchment_names = json.loads(request.read())
            callback(jam_parchment_names)

        self._make_request(jam_kind_name, inner)

    def read_texts(self, jam_kind_name: str, jam_id_name: str, start: int, end: Optional[int], *,
                   callback: Callable) -> bytes:
        assert jam_kind_pattern.fullmatch(jam_kind_name)
        assert jam_id_pattern.fullmatch(jam_id_name)
        block_size = int(jam_id_pattern.fullmatch(jam_id_name).groups()[-1])
        smart_end = end if (end is not None) else float('inf')
        assert 0 <= start <= smart_end
        def inner(request: ajax.ajax) -> None:
            conjoined_texts = request.read()
            assert isinstance(conjoined_texts, bytes)
            assert len(conjoined_texts) % block_size == 0
            texts = tuple(map(bytes.decode, sliced(conjoined_texts, block_size)))
            assert 0 <= len(texts) <= (smart_end - start)
            callback(texts)

        self._make_request(f'{jam_kind_name}/{jam_id_name}/{start}..{end}', inner, binary=True)

    def read_jams(self, jam_kind_name: str, jam_id_name: str, start: int, end: int, *,
                  callback: Callable) -> bytes:
        self.read_texts(jam_kind_name, jam_id_name, start, end,
                        callback=(lambda texts: callback(tuple(map(json.loads, texts)))))

    # def _get_jam_kinds(self) -> Sequence[JamKind]:
        # response = self._make_request('')
        # result = response.json()
        # assert isinstance(result, list)
        # assert set(map(type, result)) == {str}
        # return result

    # def _get_jam_kind_length(self, jam_kind_name: str) -> int:
        # return len(self._get_jam_parchments(jam_kind_name))

    # def _get_jam_parchment_length(self, jam_kind_name: str, jam_id: Union[str, JamId]) -> int:
        # (jam_id, jam_id_name) = JamId.parse(jam_id)
        # assert re.fullmatch('^[a-z0-9_]+$', jam_kind_name)
        # response = self._make_request(f'{jam_kind_name}/{jam_id_name}/length')
        # result = response.text
        # assert re.fullmatch('^[0-9]+$', result)
        # return int(result)








# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

'''This module defines the `animate` function that turns GridRoyale states into animation data.'''


from __future__ import annotations

from typing import (Optional, Tuple, Union, Container, Hashable, Iterator,
                    Iterable, Any, Dict, FrozenSet, Callable, Type, Sequence)
import itertools
import math

from .vectoring import Step
from .core import (State, Observation, SHOT_REWARD, COLLISION_REWARD, STARTING_REWARD,
                   NOTHING_REWARD, FOOD_REWARD, Bullet)

from marley import gamey


_directions = ((0, -1), (1, 0), (0, 1), (-1, 0))

color_map = {
    SHOT_REWARD: (0x99, 0x44, 0x44),
    COLLISION_REWARD: (0x88, 0x88, 0x22),
    STARTING_REWARD: (0x55, 0x55, 0x55),
    NOTHING_REWARD: (0x55, 0x55, 0x55),
    FOOD_REWARD: (0x55, 0x55, 0x55),
}

def animate(states: Iterable[State]) -> Iterator[dict]:
    state_iterator = iter(states)
    try:
        first_state = next(state_iterator)
    except StopIteration:
        return ()

    double_state_iterator = gamey.utils.iterate_windowed_pairs(
        itertools.chain((first_state, first_state), state_iterator)
    )

    for old_state, new_state in double_state_iterator:

        old_state: State
        new_state: State

        ############################################################################################
        ### Processing players: ####################################################################
        #                                                                                          #
        players = []

        for letter, old_observation in old_state.items():
            old_observation: Observation
            new_observation: Observation = new_state[letter]
            players.append((
                letter, (tuple(old_observation.position), tuple(new_observation.position)),
                (color_map[old_observation.reward], color_map[new_observation.reward]), (1, 1),
            ))
        #                                                                                          #
        ### Finished processing players. ###########################################################
        ############################################################################################

        ############################################################################################
        ### Processing food: #######################################################################
        #                                                                                          #
        food = []
        old_food_positions = set(map(tuple, old_state.food_positions))
        new_food_positions = set(map(tuple, new_state.food_positions))

        created_food_positions = new_food_positions - old_food_positions
        deleted_food_positions = old_food_positions - new_food_positions
        remaining_food_positions = old_food_positions & new_food_positions

        for created_food_position in created_food_positions:
            food.append((created_food_position, (0, 1)))

        for deleted_food_position in deleted_food_positions:
            food.append((deleted_food_position, (1, 0)))

        for remaining_food_position in remaining_food_positions:
            food.append((remaining_food_position, (1, 1)))

        #                                                                                          #
        ### Finished processing food. ##############################################################
        ############################################################################################

        ############################################################################################
        ### Processing bullets: ####################################################################
        #                                                                                          #
        bullets = []
        old_bullets_wip = set(itertools.chain.from_iterable(old_state.bullets.values()))
        new_bullets_wip = set(itertools.chain.from_iterable(new_state.bullets.values()))

        while old_bullets_wip:
            old_bullet: Bullet = old_bullets_wip.pop()
            desired_new_bullet = old_bullet.get_next_bullet()
            try:
                new_bullets_wip.remove(desired_new_bullet)
            except KeyError:
                bullet_hit = True
            else:
                bullet_hit = False

            bullets.append((
                (tuple(old_bullet.position), tuple(desired_new_bullet.position)),
                old_bullet.direction.angle_from_top,
                (1, 1 - bullet_hit)
            ))

        while new_bullets_wip:
            new_bullet: Bullet = new_bullets_wip.pop()
            bullets.append((
                (tuple(new_bullet.get_previous_bullet().position), tuple(new_bullet.position)),
                new_bullet.direction.angle_from_top,
                (0, 1)
            ))

        #                                                                                          #
        ### Finished processing bullets. ###########################################################
        ############################################################################################

        ############################################################################################
        ### Processing walls: ######################################################################
        #                                                                                          #
        walls_dict = {}

        for wall_position in old_state.living_wall_positions:
            walls_dict.setdefault(wall_position, [0, 0])[0] = 1

        for wall_position in old_state.destroyed_wall_positions:
            walls_dict.setdefault(wall_position, [0, 0])[0] = 0.5

        for wall_position in new_state.living_wall_positions:
            walls_dict.setdefault(wall_position, [0, 0])[1] = 1

        for wall_position in new_state.destroyed_wall_positions:
            walls_dict.setdefault(wall_position, [0, 0])[1] = 0.5

        walls = tuple(sorted((tuple(wall_position), tuple(opacities)) for wall_position, opacities
                             in walls_dict.items()))

        #                                                                                          #
        ### Finished processing walls. #############################################################
        ############################################################################################

        yield {
            'board_size': old_state.board_size,
            'players': tuple(sorted(players)),
            'food': tuple(sorted(food)),
            'bullets': tuple(sorted(bullets)),
            'walls': walls,
        }



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

'''Tools for dealing with compatibility with different Python versions.'''

import functools
from typing import Callable,  Any

# Provide functools.cache in older Python versions:
if not hasattr(functools, 'cache'):
    def cache(user_function: Callable, /) -> Any:
        '''Simple lightweight unbounded cache.  Sometimes called "memoize".'''
        return functools.lru_cache(maxsize=None)(user_function)
    functools.cache = cache



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

'''A package for exposing the GridRoyale frontend interface as a local web server.'''

from .server import *


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.
import setuptools
import re
import sys
import pathlib



def read_file(filename):
    with open(filename) as file:
        return file.read()
    
def was_arezzo_built():
    arezzo_folder = pathlib.Path(__file__).parent / 'marley' / 'arezzo' / 'arezzo'
    arezzo_src_folder = arezzo_folder / 'src'
    arezzo_dist_folder = arezzo_folder / 'dist'
    all_arezzo_src_paths = tuple(arezzo_src_folder.rglob('*'))
    all_arezzo_dist_paths = tuple(arezzo_dist_folder.rglob('*'))
    newest_arezzo_src_creation_time = max(path.stat().st_ctime for path in all_arezzo_src_paths)
    oldest_arezzo_dist_modification_time = min(path.stat().st_mtime for path in all_arezzo_dist_paths)
    return oldest_arezzo_dist_modification_time > newest_arezzo_src_creation_time

if any(('dist' in arg) for arg in sys.argv):
    if not was_arezzo_built():
        raise Exception(
            "Arezzo source code was modified, it needs to be rebuilt before you can build Marley. "
            "Rebuild it by going into the `marley/arezzo/arezzo` folder and running `npm run "
            "build`"
        )


version = re.search("__version__ = '([0-9.]*)'",
                    read_file('marley/__init__.py')).group(1)

setuptools.setup(
    name='marley',
    version=version,
    author='Ram Rachum',
    author_email='ram@rachum.com',
    description='Marley - A framework for multi-agent reinforcment learning',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/cool-RR/marley',
    packages=setuptools.find_packages(exclude=['tests*']),
    install_requires=read_file('requirements.txt'),
    include_package_data=True, # For including the frontend files
    python_requires='>=3.8',
    extras_require={
        'tests': {
            'pytest',
            'pytest-xdist',
            'pytest-html',
        },
    },
    entry_points={
        'console_scripts': [
            'marley = marley.commanding:marley'
        ],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Artificial Life',
    ],
)



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import operator as operator_module
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar)
import collections.abc
import itertools
import functools

_initial_missing = object()


def union(items: Iterable[Any], initial: Any = _initial_missing) -> Any:
    extra_args = (initial,) if initial is not _initial_missing else ()
    return functools.reduce(operator_module.or_, items, *extra_args)

def intersection(items: Iterable[Any], initial: Any = _initial_missing) -> Any:
    extra_args = (initial,) if initial is not _initial_missing else ()
    return functools.reduce(operator_module.and_, items, *extra_args)


class ChangeTracker:
    def __init__(self):
        self.value = object()

    def __call__(self, value: Any) -> bool:
        try:
            return (self.value != value)
        finally:
            self.value = value





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import time
import threading
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict, Set)
from types import TracebackType
import concurrent.futures
from bidict import bidict
import logging

import more_itertools

from .weighting import Weight, CalfWeight, TadpoleWeight
from .gaining import ThinGain, FatGain
from .jobbing import Job, ThinJob, FatJob
from .antillesing import Antilles
from . import utils

logger = logging.getLogger(__name__)


DEFAULT_MAX_FUTURES = 100

class CuteExecutorFull(Exception):
    pass

def worker_setup_then_function(worker_setup_function: Optional[callable],
                               worker_setup_args: Sequence, function: Callable) -> Any:
    if worker_setup_function is not None:
        worker_setup_function(*worker_setup_args)
    return function()



class CuteExecutorMixin(concurrent.futures.Executor):
    def __init__(self, *, max_futures: int = DEFAULT_MAX_FUTURES,
                 worker_setup_function: Optional[Callable] = None,
                 worker_setup_args: Sequence = ()) -> None:
        self.max_futures = max_futures
        self.futures = set()
        self.worker_setup_function = worker_setup_function
        self.worker_setup_args = worker_setup_args

    def can_accept_more_futures(self) -> bool:
        return len(self.futures) < self.max_futures

    def submit(self, function: Callable) -> concurrent.futures.Future:
        if not self.can_accept_more_futures():
            raise CuteExecutorFull
        else:
            future = super().submit(worker_setup_then_function, self.worker_setup_function,
                                    self.worker_setup_args, function)
            self.futures.add(future)
            return future

    def _get_completed_futures(self) -> Set[concurrent.futures.Future]:
        iterator = concurrent.futures.as_completed(self.futures, timeout=0)
        return set(
            more_itertools.iter_except(
                iterator.__next__,
                (StopIteration, concurrent.futures.TimeoutError)
            )
        )

    def get_completed_futures(self, *, wait_for_at_least_one_if_non_empty: bool = False) -> \
                                                                     Set[concurrent.futures.Future]:
        completed_futures = self._get_completed_futures()
        if wait_for_at_least_one_if_non_empty and not completed_futures and self.futures:
            logger.debug(f"There are no completed futures, and there are {len(self.futures)} "
                         f"pending futures. We were asked to wait in this scenario for just one "
                         f"future, so we'll wait.")
            completed_futures = {more_itertools.first(
                                                     concurrent.futures.as_completed(self.futures))}

        for future in completed_futures:
            future: concurrent.futures.Future
            future.result() # If the future got an exception, raise it here.
            self.futures.remove(future)
        return completed_futures

    # def iterate_futures_as_completed(self) -> Iterator[concurrent.futures.Future]:
        # for future in concurrent.futures.as_completed(self.futures):
            # self.futures.remove(future)
            # yield future


class CuteProcessExecutor(CuteExecutorMixin, concurrent.futures.ProcessPoolExecutor):
    def __init__(self, max_workers=None, *, max_futures: int = DEFAULT_MAX_FUTURES,
                 worker_setup_function: Optional[Callable] = None,
                 worker_setup_args: Sequence = ()) -> None:
        CuteExecutorMixin.__init__(self, max_futures=max_futures,
                                   worker_setup_function=worker_setup_function,
                                   worker_setup_args=worker_setup_args)
        concurrent.futures.ProcessPoolExecutor.__init__(self, max_workers=max_workers)
        for _ in range(2 * max_workers):
            concurrent.futures.ProcessPoolExecutor.submit(self, sum, (),)

class CuteThreadExecutor(CuteExecutorMixin, concurrent.futures.ThreadPoolExecutor):
    def __init__(self, max_workers=None, *, max_futures: int = DEFAULT_MAX_FUTURES,
                 worker_setup_function: Optional[Callable] = None,
                 worker_setup_args: Sequence = ()) -> None:
        CuteExecutorMixin.__init__(self, max_futures=max_futures,
                                   worker_setup_function=worker_setup_function,
                                   worker_setup_args=worker_setup_args)
        concurrent.futures.ThreadPoolExecutor.__init__(self, max_workers=max_workers)



# Copyright 2020 Ram Rachum.
# This program is distributed under the MIT license.

from __future__ import generator_stop
from __future__ import annotations

import collections.abc
import functools
import threading
import itertools

from . import iterating

infinity = float('inf')

class _SENTINEL:
    '''Sentinel used to detect the end of an iterable.'''


def _convert_index_to_exhaustion_point(index):
    '''
    Convert an index to an "exhaustion point".

    The index may be either an integer or infinity.

    "Exhaustion point" means "until which index do we need to exhaust the
    internal iterator." If an index of `3` was requested, we need to exhaust it
    to index `3`, but if `-7` was requested, we have no choice but to exhaust
    the iterator completely (i.e. to `infinity`, actually the last element,)
    because only then we could know which member is the seventh-to-last.
    '''
    assert isinstance(index, int) or index == infinity
    if index >= 0:
        return index
    else: # i < 0
        return infinity


def _with_lock(method):
    '''Decorator for using the `LazyTuple`'s lock.'''
    @functools.wraps(method)
    def inner(self, *args, **kwargs):
        with self.lock:
            return method(*args, **kwargs)
    return inner


@functools.total_ordering
class LazyTuple(collections.abc.Sequence):
    '''
    A lazy tuple which requests as few values as possible from its iterator.

    Wrap your iterators with `LazyTuple` and enjoy tuple-ish features like
    indexed access, comparisons, length measuring, element counting and more.

    Example:

        def my_generator():
            yield from ('hello', 'world', 'have', 'fun')

        lazy_tuple = LazyTuple(my_generator())

        assert lazy_tuple[2] == 'have'
        assert len(lazy_tuple) == 4

    `LazyTuple` holds the given iterable and pulls items out of it. It pulls as
    few items as it possibly can. For example, if you ask for the third
    element, it will pull exactly three elements and then return the third one.

    Some actions require exhausting the entire iterator. For example, checking
    the `LazyTuple` length, or doing indexex access with a negative index.
    (e.g. asking for the seventh-to-last element.)

    If you're passing in an iterator you definitely know to be infinite,
    specify `definitely_infinite=True`.
    '''

    def __init__(self, iterable, *, definitely_infinite=False):
        was_given_a_sequence = isinstance(iterable, collections.abc.Sequence) \
                                        and not isinstance(iterable, LazyTuple)

        self.is_exhausted = True if was_given_a_sequence else False
        '''Flag saying whether the internal iterator is tobag exhausted.'''

        self.collected_data = iterable if was_given_a_sequence else []
        '''All the items that were collected from the iterable.'''

        self._iterator = None if was_given_a_sequence else iter(iterable)
        '''The internal iterator from which we get data.'''

        self.definitely_infinite = definitely_infinite
        '''
        The iterator is definitely infinite.

        The iterator might still be infinite if this is `False`, but if it's
        `True` then it's definitely infinite.
        '''

        self.lock = threading.Lock()
        '''Lock used while exhausting to make `LazyTuple` thread-safe.'''


    @classmethod
    def factory(cls, *, definitely_infinite=False):
        '''
        Decorator to make generators return a `LazyTuple`.

        Example:

            @LazyTuple.factory()
            def my_generator():
                yield from ['hello', 'world', 'have', 'fun']

        This works on any function that returns an iterator. todo: Make it work
        on iterator classes.
        '''
        def decorator(function):
            @functools.wraps(function)
            def inner(*args, **kwargs):
                return cls(function(*args, **kwargs),
                           definitely_infinite=definitely_infinite)
            return inner
        return decorator


    @property
    def known_length(self):
        '''
        The number of items which have been taken from the internal iterator.
        '''
        return len(self.collected_data)


    def exhaust(self, i=infinity):
        '''
        Take items from the internal iterators and save them.

        This will take enough items so we will have `i` items in total,
        including the items we had before.
        '''
        if self.is_exhausted:
            return

        elif isinstance(i, int) or i == infinity:
            exhaustion_point = _convert_index_to_exhaustion_point(i)

        else:
            assert isinstance(i, slice)
            raise NotImplementedError

            # # todo: can be smart and figure out if it's an empty slice and then
            # # not exhaust.

            # canonical_slice = sequence_tools.CanonicalSlice(i)

            # exhaustion_point = max(
                # _convert_index_to_exhaustion_point(canonical_slice.start),
                # _convert_index_to_exhaustion_point(canonical_slice.stop)
            # )

            # if canonical_slice.step > 0: # Compensating for excluded last item:
                # exhaustion_point -= 1

        while len(self.collected_data) <= exhaustion_point:
            try:
                with self.lock:
                    self.collected_data.append(next(self._iterator))
            except StopIteration:
                self.is_exhausted = True
                break


    def __getitem__(self, i):
        '''Get item by index, either an integer index or a slice.'''
        self.exhaust(i)
        result = self.collected_data[i]
        if isinstance(i, slice):
            return tuple(result)
        else:
            return result


    def __len__(self):
        if self.definitely_infinite:
            return 0 # Unfortunately infinity isn't supported.
        else:
            self.exhaust()
            return len(self.collected_data)


    def __eq__(self, other):
        if not isinstance(other, LazyTuple):
            return False
        for i, j in itertools.zip_longest(self, other, fillvalue=_SENTINEL):
            if (i is _SENTINEL) or (j is _SENTINEL):
                return False
            if i != j:
                return False
        return True


    def __ne__(self, other):
        return not self.__eq__(other)


    def __bool__(self):
        try: next(iter(self))
        except StopIteration: return False
        else: return True


    def __lt__(self, other):
        if not self and other:
            return True
        elif self and not other:
            return False
        elif not self and not other:
            return False
        for a, b in itertools.zip_longest(self, other,
                                           fillvalue=_SENTINEL):
            if a is _SENTINEL:
                # `self` ran out. Now there can be two cases: (a) `other` ran
                # out too or (b) `other` didn't run out yet. In case of (a), we
                # have `self == other`, and in case of (b), we have `self <
                # other`. In any case, `self <= other is True` so we can
                # unconditionally return `True`.
                return True
            elif b is _SENTINEL:
                assert a is not _SENTINEL
                return False
            elif a == b:
                continue
            elif a < b:
                return True
            else:
                assert a > b
                return False


    def __repr__(self):
        '''
        Return a human-readeable representation of the `LazyTuple`.

        Example:

            <LazyTuple: (1, 2, 3, ...)>

        The '...' denotes a non-exhausted lazy tuple.
        '''
        if self.is_exhausted:
            inner = repr(self.collected_data)

        else: # not self.exhausted
            if self.collected_data == []:
                inner = '(...)'
            else:
                inner = '%s...' % repr(self.collected_data)
        return '<%s: %s>' % (self.__class__.__name__, inner)


    def __add__(self, other):
        return tuple(self) + tuple(other)


    def __radd__(self, other):
        return tuple(other) + tuple(self)


    def __mul__(self, other):
        return tuple(self).__mul__(other)


    def __rmul__(self, other):
        return tuple(self).__rmul__(other)


    def __hash__(self):
        '''
        Get the `LazyTuple`'s hash.

        Note: Hashing the `LazyTuple` will completely exhaust it.
        '''
        if self.definitely_infinite:
            raise TypeError("An infinite `LazyTuple` isn't hashable.")
        else:
            self.exhaust()
            return hash(tuple(self))


def massage_iterable(iterable, unallowed_types=(bytes,)):
    '''
    Return a version of `iterable` that is an immutable sequence.

    If `iterable` is already an immutable sequence, it returns it as is; otherwise, it makes it
    into a `LazyTuple`.
    '''
    assert isinstance(iterable, collections.abc.Iterable)
    if isinstance(iterable, (collections.abc.MutableSequence, *unallowed_types)) or \
       not isinstance(iterable, collections.abc.Sequence):
        return LazyTuple(iterable)
    else:
        return iterable



class ChainSpace(collections.abc.Sequence):
    '''
    A space of sequences chained together.

    This is similar to `itertools.chain`, except that items can be fetched by
    index number rather than just iteration.

    Example:

        >>> chain_space = ChainSpace(('abc', (1, 2, 3)))
        >>> chain_space
        <ChainSpace: 3+3>
        >>> chain_space[4]
        2
        >>> tuple(chain_space)
        ('a', 'b', 'c', 1, 2, 3)
        >>> chain_space.index(2)
        4

    '''
    def __init__(self, iterables):
        self.sequences = LazyTuple(map(massage_iterable, iterables))
        self.accumulated_lengths = LazyTuple(
            itertools.accumulate(map(len, self.sequences), initial=0)
        )
        self.accumulated_lengths.exhaust(1)

    def __len__(self):
        return self.accumulated_lengths[-1]

    known_length = property(lambda self: self.accumulated_lengths.collected_data[-1])

    def __repr__(self):
        if self.sequences.is_exhausted:
            text = f'Exhausted; {len(self.sequences)} sequences; {len(self)} items'
        else:
            text = (f'Not exhausted; >={len(self.sequences.collected_data)} sequences; '
                    f'>={self.known_length} items')
        return f'<{type(self).__name__}: {text}>'

    def __getitem__(self, i):
        if isinstance(i, slice):
            raise NotImplementedError
        assert isinstance(i, int)
        if i <= -1:
            fixed_i += len(self) # Exhausting the chain space.
        else:
            fixed_i = i
        if fixed_i < 0:
            raise IndexError(i)
        required_length = fixed_i + 1
        if self.accumulated_lengths.is_exhausted and required_length > len(self):
            raise IndexError

        iterator = zip(self.sequences, iterating.iterate_windowed_pairs(self.accumulated_lengths))
        for sequence, (previous_accumulated_length, accumulated_length) in iterator:
            if accumulated_length >= required_length:
                assert previous_accumulated_length < required_length
                return sequence[i - previous_accumulated_length]
        else:
            raise IndexError(i)

    def __iter__(self):
        return itertools.chain.from_iterable(self.sequences)

    _reduced = property(lambda self: (type(self), self.sequences))

    __eq__ = lambda self, other: (isinstance(other, ChainSpace) and
                                  self._reduced == other._reduced)

    def __contains__(self, item):
        return any(item in sequence for sequence in self.sequences
                   if (not isinstance(sequence, str) or isinstance(item, str)))

    def index(self, item):
        '''Get the index number of `item` in this space.'''
        for sequence, accumulated_length in zip(self.sequences,
                                                self.accumulated_lengths):
            try:
                index_in_sequence = sequence.index(item)
            except ValueError:
                pass
            except TypeError:
                assert isinstance(sequence, (str, bytes)) and \
                                           (not isinstance(item, (str, bytes)))
            else:
                return index_in_sequence + accumulated_length
        else:
            raise ValueError

    def __bool__(self):
        try: next(iter(self))
        except StopIteration: return False
        else: return True




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import random
import weakref
import concurrent.futures
import numbers
import functools
import collections.abc
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Mapping, MutableMapping)
import hashlib
import weakref
import dataclasses
import pathlib
import tempfile

import more_itertools
import numpy as np

from marley.gamey import GameySwankDatabase


class MarleySwankDatabase(GameySwankDatabase):
    def get_swank_types(self) -> dict:
        return super().get_swank_types() | set()






# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import generator_stop

import numpy as np
import pathlib
import collections
import dataclasses
from typing import Optional, Union, Tuple, Iterable, Iterator, Sequence
import numbers
import random
import tempfile

import more_itertools
from .iterating import *
from .weak_key_identity_dict import WeakKeyIdentityDict
from .chain_space import ChainSpace


def cute_div(x: numbers.Number, y: numbers.Number) -> numbers.Number:
    '''Divide x by y, allowing to divide by 0 and giving infinity with the right sign.'''
    try:
        return x / y
    except ZeroDivisionError:
        if x == 0:
            raise
        else:
            return x * float('inf')

def clamp(number, /, minimum, maximum):
    if number < minimum:
        return minimum
    elif number > maximum:
        return maximum
    else:
        return number


class NiceDataclass(collections.abc.Sequence):
    __len__ = lambda self: len(dataclasses.fields(self))
    __iter__ = lambda self: map(
        self.__dict__.__getitem__,
        (field.name for field in dataclasses.fields(self))
    )
    __getitem__ = lambda self, i: tuple(self)[i]

def is_structured_array(array: np.ndarray) -> bool:
    return isinstance(array.dtype, np.dtype) and len(array.dtype) >= 1

def shuffled(sequence: Sequence) -> list:
    '''
    Return a list with all the items from `sequence` shuffled.

    Example:

        >>> random_tools.shuffled([0, 1, 2, 3, 4, 5])
        [0, 3, 5, 1, 4, 2]

    '''
    sequence_copy = list(sequence)
    random.shuffle(sequence_copy)
    return sequence_copy





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import math
import inspect
import re
import abc
import random
import itertools
import collections.abc
import statistics
import concurrent.futures
import enum
import functools
import numbers
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar)
import dataclasses

import more_itertools
import numpy as np

from marley.utils import ImmutableDict
from marley.jamswank.swanking import SimpleField, SavvyField, SwankField
from marley.gamey.gamey_swanking import GameySwank
from . import utils
from . import exceptions



class _ActionType(abc.ABCMeta):# collections.abc.Sequence):
    __iter__ = lambda cls: iter(cls.all_actions)
    __len__ = lambda cls: len(cls.all_actions)
    def __getitem__(cls, i: int):
        if i >= len(cls):
            raise IndexError
        for j, item in enumerate(cls):
            if j == i:
                return cls
        raise RuntimeError

    @property
    def n_neurons(cls) -> int:
        try:
            return cls._n_neurons
        except AttributeError:
            cls._n_neurons = len(cls)
            return cls._n_neurons



_action_regex_head = re.compile(r'[A-Za-z0-9.]')
_action_regex_tail = re.compile(r'[A-Za-z0-9_.\-/>]*')
_action_regex = re.compile(f'^{_action_regex_head.pattern}'
                           f'{_action_regex_tail.pattern}$')

@functools.total_ordering
class Action(metaclass=_ActionType):
    all_actions: Sequence[Action]
    n_neurons: int

    def __lt__(self, other):
        return self.all_actions.index(self) < self.all_actions.index(other)

    def slugify(self) -> str:
        raw = str(self)
        first_letter = raw[0]
        prefix = '' if _action_regex_head.fullmatch(first_letter) else '0'
        characters = ((c if _action_regex_tail.fullmatch(c) else '-') for c in raw)
        result = f'{prefix}{"".join(characters)}'
        assert _action_regex.fullmatch(result)
        return result

    def to_neural(self) -> np.ndarray:
        # Implementation for simple discrete actions. Can override.
        try:
            return self._to_neural
        except AttributeError:
            self._to_neural = np.array([(self == action) for action in type(self)],
                                        dtype=bool)
            return self._to_neural

    @classmethod
    def from_neural(cls, neural: Iterable) -> Action:
        # Implementation for simple discrete actions. Can override.
        return cls[tuple(neural).index(1)]


class _ActionEnumType(type(Action), type(enum.Enum)):
    pass


class ActionEnum(Action, enum.Enum, metaclass=_ActionEnumType):
    # todo: use in Blackjack
    pass


class Observation(abc.ABC):
    state: State
    legal_actions: Tuple[Action, ...]
    neural_dtype: np.dtype

    @property
    def neural(self) -> np.ndarray:
        '''A structured array of numbers that represents the observation for a neural network.'''
        try:
            return self._neural
        except AttributeError:
            self._neural = self._to_neural()
            return self._neural

    @abc.abstractmethod
    def _to_neural(self) -> np.ndarray:
        '''Represent the observation as a structured array of numbers for a neural network.'''
        pass

    @abc.abstractmethod
    def to_savvy_content(self) -> tuple:
        pass



PlayerId = TypeVar('PlayerId', bound=Hashable)


@dataclasses.dataclass(order=True, frozen=True)
class Story:
    old_observation: Observation
    action: Action
    reward: numbers.Number
    new_observation: Observation


from .aggregating import State


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import operator as operator_module
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar)
import collections.abc
import itertools
import functools

from .pointing import Point
from .point_containing import PointContainer, ArithmeticOther


class IntervalType(type(PointContainer)):
    def __getitem__(cls, arg) -> Interval:
        as_tuple = arg if isinstance(arg, tuple) else (arg,)
        return Interval(as_tuple)

def slice_to_range(s: slice) -> range:
    assert s.step is None
    return range(s.start, s.stop)


@functools.total_ordering
class Interval(PointContainer, metaclass=IntervalType):

    def __init__(self, numbers_and_slices: Iterable[Union[int, tuple[int, int]]]) -> None:
        processed_numbers_and_slices = []
        numbers_and_slices_tuple = tuple(numbers_and_slices)
        for item in numbers_and_slices_tuple:
            if isinstance(item, int):
                assert item >= 0
                processed_numbers_and_slices.append(item)
            else:
                if isinstance(item, tuple):
                    (start, end) = item
                else:
                    assert isinstance(item, slice)
                    start = item.start
                    end = item.stop

                assert isinstance(start, int)
                assert isinstance(end, int)
                assert 0 <= start <= end
                if start <= end - 2:
                    processed_numbers_and_slices.append(slice(start, end))
                elif start == end - 1:
                    processed_numbers_and_slices.append(start)
                else:
                    assert start == end
                    processed_numbers_and_slices = [slice(0, 0)] * len(numbers_and_slices_tuple)
                    break

        self.numbers_and_slices = tuple(processed_numbers_and_slices)
        self.dimensions = len(self.numbers_and_slices)
        self.sliceful_tuple = tuple((slice(item, item + 1) if isinstance(item, int) else item)
                                    for item in self.numbers_and_slices)
        self.tupleful_tuple = tuple((s.start, s.stop) for s in self.sliceful_tuple)
        self.intervals = (self,)

    def as_tuple(self) -> tuple:
        return self.numbers_and_slices

    def __iter__(self) -> Iterable[Point]:
        return map(Point, itertools.product(*map(slice_to_range, self.sliceful_tuple)))

    def _reduce(self) -> tuple:
        return (type(self), repr(self))

    def __repr__(self) -> str:
        if self.dimensions == 0:
            assert self.numbers_and_slices == ()
            return f'{type(self).__name__}(())'
        else:
            foo = ', '.join(
                str(item) if isinstance(item, int) else f'{item.start}:{item.stop}'
                for item in self.numbers_and_slices
            )
            return f'{type(self).__name__}[{foo}]'

    def __lt__(self, other: Interval) -> bool:
        assert isinstance(other, Interval)
        if self.dimensions != other.dimensions:
            return (self.dimensions < other.dimensions)
        else:
            return (self.sliceful_tuple < other.sliceful_tuple)

    def to_savvy_content(self):
        return (self.tupleful_tuple,)



from .int_crowding import IntCrowd
from .functions import intervals_to_points, points_to_intervals, canonicalize_intervals


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import operator as operator_module
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar)
import collections.abc
import itertools
import functools


@functools.total_ordering
class Point(collections.abc.Sequence):
    def __init__(self, numbers: Iterable[int]) -> None:
        self.numbers = tuple(numbers)
        for number in self.numbers:
            assert isinstance(number, int)
            assert number >= 0
        self.dimensions = len(self.numbers)

    def __getitem__(self, i: int) -> int:
        return self.numbers[i]

    def __len__(self) -> int:
        return len(self.numbers)

    def __iter__(self) -> Iterable[int]:
        return iter(self.numbers)

    def __reversed__(self) -> Iterable[int]:
        return reversed(self.numbers)

    def as_tuple(self) -> tuple:
        return self.numbers

    def as_interval(self) -> Interval:
        return Interval(self)

    def __lt__(self, other: Point) -> bool:
        assert isinstance(other, Point)
        if self.dimensions != other.dimensions:
            return (self.dimensions < other.dimensions)
        else:
            return (self.numbers < other.numbers)

    def __eq__(self, other: Point) -> bool:
        assert isinstance(other, Point)
        return self.numbers == other.numbers

    def __hash__(self) -> int:
        return hash((type(self), self.numbers))

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.numbers})'



from .intervaling import Interval